#include "global.hpp"
#include "input.hpp"
#include "updater.hpp"

Playback::Playback(const std::string& n)
  : enabled(false), playback_name(n)
{}

void Playback::add_command(const GlobalCommand& c)
{
  commands.push_back(c);
}

bool Playback::check_commands(const Input& input) const
{
  bool status = false;

  for (std::vector<GlobalCommand>::const_iterator i = commands.begin(), 
	 end = commands.end(); i != end; ++i)
    if (i->check_func_base(input)) {
      i->call_back_f->call();
      status = true;
      break;
    }

  S_ScreenUpdater::get_instance()->trigger.trigger_input(input.command);

  return status;
}

void Playback::enable()
{
  enabled = true;
}

void Playback::disable()
{
  enabled = false;
}

std::string Playback::name() const
{
  return playback_name;
}

void Playback::clean_up()
{
  for (std::vector<GlobalCommand>::const_iterator i = commands.begin(), 
	 end = commands.end(); i != end; ++i)
    delete i->call_back_f;
}

void Global::clean_up()
{
  for (std::vector<GlobalCommand>::const_iterator i = commands.begin(), 
	 end = commands.end(); i != end; ++i) {
    if (i->check_func != 0)
      delete i->check_func;

    delete i->call_back_f;
  }

  for (std::vector<Playback>::iterator i = playbacks.begin(), 
	 end = playbacks.end(); i != end; ++i)
    i->clean_up();

  delete reconfigure_callback;
  delete lcd_print_output_func;
  delete movie_playback_func;
}

bool GlobalCommandBase::check_func_base(const Input& i) const
{
  return (i.command == command);
}

void Global::add_command(const GlobalCommand& c)
{
  commands.push_back(c);
}

bool Global::check_commands(const Input& input, bool disable_mode_change)
{
  for (std::vector<GlobalCommand>::const_iterator i = commands.begin(), 
	 end = commands.end(); i != end; ++i) {
    bool ok = false;

    if (i->check_func != 0)
      ok = i->check_func->call(input, disable_mode_change);
    else
      ok = i->check_func_base(input);

    if (ok)
      i->call_back_f->call();
  }

  for (std::vector<Playback>::const_iterator i = playbacks.begin(), 
	 end = playbacks.end(); i != end; ++i)
    if (i->name() == playback_name)
      return i->check_commands(input);

  return false;
}

void Global::add_playback(const Playback& p)
{
  playbacks.push_back(p);
}

bool Global::set_playback(const std::string& name)
{
  for (std::vector<Playback>::const_iterator i = playbacks.begin(), 
	 end = playbacks.end(); i != end; ++i) {
    if (i->name() == name) {
      playback_name = name;
      return true;
    }
  }
  return false;
}

bool Global::is_playback(const std::string& name)
{
  return name == playback_name;
}

void Global::set_playback_offset(int off)
{
  offset = off;
}

int Global::get_playback_offset()
{
  return offset;
}

void Global::reconfigure_playback()
{
  if (reconfigure_callback)
    reconfigure_callback->call();
  else
    std::cerr << "ERROR: no playback reconfigure function bound" << std::endl;
}

void Global::register_reconfigure(TimeFunctorv<AudioPlayer>* callback)
{
  reconfigure_callback = callback;
}

void Global::stop_bit_input_check(const Input& i)
{
  if (i.mode == "general" && i.command == "cancel")
    cancel = true;
}

void Global::reset_stop_bit()
{
  cancel = false;
}

bool Global::check_stop_bit()
{
  bool tmp = cancel;
  cancel = false;
  return tmp;
}

void Global::set_valid_keys(const std::string& keys)
{
  valid_keys = keys;
}

std::string Global::get_valid_keys()
{
  return valid_keys;
}

// LCD

void Global::register_lcd_output_func(ConstFunctor<LcdPrint, std::string, void>* lcd_output)
{
  lcd_output_func = lcd_output;
}

void Global::register_lcd_print_output_func(TimeFunctorv<LcdPrint>* lcd_print_output)
{
  lcd_print_output_func = lcd_print_output;
  registered_lcd_output = true;
}

bool Global::lcd_output_possible()
{
  return registered_lcd_output;
}

void Global::lcd_add_output(const std::string& message)
{
  lcd_output_func->call(message);
}

void Global::lcd_print()
{
  lcd_print_output_func->call();
}

// Movie

bool Global::movie_playback_possible()
{
  return registered_movie_playback;
}

void Global::register_movie_playback_func(ConstFunctor<Movie, std::string, void>* movie_playback)
{
  movie_playback_func = movie_playback;
  registered_movie_playback = true;
}

void Global::play_movie(const std::string& path)
{
  movie_playback_func->call(path);
}

void Global::add_second_menu_item(const ExtraMenuItem& e)
{
  menu_items.push_back(e);
}

Global::Global()
  : cancel(false), playback_name(""), registered_lcd_output(false), registered_movie_playback(false)
{
  InputMaster *input_master = S_InputMaster::get_instance();
  input_master->set_map("default");
  add_second_menu_item(ExtraMenuItem(gettext("Playback control"), 
				     input_master->find_shortcut("playback_commands"), 
				     new TimeFunctorv<Global>(this, &Global::playlist_commands)));
}

void Global::playlist_commands()
{
  if (playback_name.empty())
    return;

  ExtraMenu em = ExtraMenu(gettext("Playlist Commands"));

  InputMaster *input_master = S_InputMaster::get_instance();
  
  for (std::vector<Playback>::const_iterator i = playbacks.begin(), 
	 end = playbacks.end(); i != end; ++i) {
    if (i->name() == playback_name) {
      for (std::vector<GlobalCommand>::const_iterator global_command_iter = i->commands.begin(), 
	     global_command_iend = i->commands.end(); global_command_iter != global_command_iend; ++global_command_iter) {
	em.add_persistent_item(ExtraMenuItem(global_command_iter->name, input_master->find_shortcut(global_command_iter->command), 
					     global_command_iter->call_back_f));
      }
      break;
    }
  }
  em.mainloop();
}

