#include "config.h"

#include "config.hpp"
#include "common.hpp"
#include "startmenu.hpp"
#include "updater.hpp"
#include "theme.hpp"
#include "print.hpp"

#include "plugins.hpp"

#include "renderer.hpp"
#include "input.hpp"

/* SIGHUP */
#include <signal.h>

#include <unistd.h>
#include <sstream>

#include <iostream>

// --version
#include <getopt.h>

#include "gettext.hpp"

#ifdef use_nls
#include <langinfo.h>
#include <locale.h>
#endif

using std::vector;
using std::string;

void start()
{
  //Print the info and wait for input
  for (string output = ""; output != "quit";)
    output = S_Startmenu::get_instance()->mainloop();
}

void print_options_header() 
{
  Print print(gettext("mms options:") + string("\n"), Print::INFO);
}

void print_options_footer() 
{
  Print print(Print::INFO);
  print.add_line("");
  print.add_line(gettext("<relative> : filename relative to config directory"));
  print.add_line(gettext("<absolute> : full path filename"));
  print.print();
}

void print_output_options() 
{
  Plugins *plugins = S_Plugins::get_instance();
  string output = gettext(" -o : output device [");

  for (vector<OutputPlugin*>::const_iterator iter=plugins->op_data.begin();
       iter!=plugins->op_data.end(); iter++) {
    output += (*iter)->plugin_name();
    if (*iter != plugins->op_data.back())
      output += ",";
  }
  
  output += "]";
  
  Print print(output, Print::INFO);
}

void print_input_options() 
{
  Plugins *plugins = S_Plugins::get_instance();
  string output = gettext(" -i : input device [");

  for (vector<InputPlugin*>::const_iterator iter=plugins->ip_data.begin();
       iter!=plugins->ip_data.end(); iter++) {
    output += (*iter)->plugin_name();
    if (*iter != plugins->ip_data.back())
      output += ",";
  }
  
  output += "]";
  
  Print print(output, Print::INFO);
}

void print_misc_options()
{
  string output = gettext(" -d : daemonize (run in background)") + string("\n") + gettext(" -c <absolute> : use alternate config directory");
  Print print (output, Print::INFO);
}

void print_replay_options()
{
  string output = gettext(" -r <relative> : use replay file as input (requires replay as input device)") + string("\n") + 
    gettext(" -b : benchmark mms by replaying as fast as possible (requires replay as input device) ") + string("\n") + 
    gettext(" -w <relative> : write replay to specified file");
    
  Print print (output, Print::INFO);
}

static struct option longopts[] =
{
  { "conf", 1, 0, 'c' },
  { "daemon", 0, 0, 'd' },
  { "replay-read", 1, 0, 'r' },
  { "benchmark", 0, 0, 'b' },
  { "replay-write", 1, 0, 'w' },
  { "version", 0, 0, 'v' },
  { "help", 0, 0, 'h' },
};

int main(int argc,char* argv[])
{
  signal(SIGINT, clean_up);
  signal(SIGTERM, clean_up);
  signal(SIGHUP, clean_up);
  signal(SIGKILL, clean_up);

  Config *conf = S_Config::get_instance();

  // overwriting variables
  string homedir = "";
  bool benchmark = false;
  string replay_read = "", replay_write = "", outdev = "", indev = "";

#ifdef use_nls
  // gettext
  setlocale(LC_ALL, ""); /* set from the environment variables */
  bindtextdomain("mms", mms_prefix"/share/locale");
  bind_textdomain_codeset("mms", nl_langinfo(CODESET));
  textdomain("mms");
#endif

  int c;
  
  while ((c = getopt_long(argc, argv, "abc:dfhlno:w:r:i:t:O:vV",
			  longopts, NULL)) != -1)
    switch (c) {

    case 'b':
      benchmark = true;
      break;
    case 'c': // config file
      homedir = optarg;
      break;
    case 'd': // daemonize
      switch (fork()) {
      case 0: break; // my child will go on
      case -1:
	fprintf(stderr,"Error forking\n");
	return -1;
      default:
	_exit(0);
      }
      if (setsid() < 0) // we won't get SIGHUP after this
	return -1;
      switch (fork()) {
      case 0: break; // my child will go on
      case -1: // error
	fprintf(stderr,"Error forking\n");
	return -1;
      default: _exit(0); // the parent process exits
      }
      break;
    case 'h': 
      print_options_header();
      print_misc_options();
      print_replay_options();
      print_input_options();
      print_output_options();
      print_options_footer();
      exit(0);
      break;
    case 'r':
      replay_read = optarg;
      break;
    case 'w':
      replay_write = optarg;
      break;
    case 'o':
      if (!strcmp(optarg,"h") || !strcmp(optarg,"help")) {
	print_options_header();
	print_output_options();
	exit(0);
      }
      outdev = optarg;
      break;
    case 'i':
      if (!strcmp(optarg,"h") || !strcmp(optarg,"help")) {
	print_options_header();
	print_input_options();
	exit(0);
      }
      indev = optarg;

      break;
    case 'v':		/* version */
    case 'V':		/* version */
      {
	DebugPrint perror("My Media System: 1.1.0-dev",
			  Print::INFO, DebugPrint::INFO);
      }
      exit (0);
    }

  // check if ~/.mms/ is available, if so use it when possible
  if (homedir.empty()) {
    homedir = getenv("HOME");
    if (file_exists(homedir + "/.mms/")) {
      conf->s_homedir(homedir + "/.mms/");
      conf->s_var_data_dir(homedir + "/.mms/");
    } else {
      conf->s_homedir("/etc/mms/");
      conf->s_var_data_dir("/var/lib/mms/");
    }
  }
  else {
    if (homedir[homedir.size()-1] != '/')
      homedir += '/';
    
    conf->s_homedir(homedir);
    conf->s_var_data_dir(homedir);
  }

  // now pass the configuration file using the correct path
  conf->parse_configuration_file();

  // set options from configuration line to overwrite config
  if (benchmark)
    conf->s_benchmark(true);

  if (!replay_read.empty())
      conf->s_replay_read(replay_read);

  if (!replay_write.empty())
      conf->s_replay_write(replay_write);

  if (!outdev.empty())
      conf->s_outdev(outdev);

  if (!indev.empty())
      conf->s_input(indev);
  
  Print print(gettext("Using ") + conf->p_homedir() + gettext(" as configuration directory"), Print::INFO);

  // init the theme engine before plugins, so that plugins can use themes
  S_Theme::get_instance();

  Plugins *plugins = S_Plugins::get_instance();
  plugins->scan();

  // Init the i/o classes
  // First init renderer and InputMaster, because I/O device plugins need them.
  // FIXME: new debug call
#ifdef mms_debug
  cerr << "Init Renderer" << endl;
#endif
  Render *render = S_Render::get_instance();
#ifdef mms_debug
  cerr << "Init input master" << endl;
#endif
  
  // make sure the temporary directory is available
  string temp_loc = conf->p_temp_image_loc();
  if (temp_loc[temp_loc.size()-1] != '/')
    temp_loc += '/';
  
  if (!file_exists(temp_loc))
    create_if_not_exists(temp_loc);

  render->set_theme();

#define mms_debug 1 

  DebugPrint pinfo_detect(gettext("Detecting input and output plugins"), Print::DEBUGGING, DebugPrint::INFO, "MMS");
  // Derefer fail to report both input & output plugins pre-errors.
  int init_error = 0;

  if (plugins->ip_data.size() == 0) {
    DebugPrint perror(gettext("No Input plugin found"),
		      Print::DEBUGGING, DebugPrint::INFO, "MMS");
    init_error++;
  }

  if (plugins->op_data.size() == 0) {
    DebugPrint perror(gettext("No output device specified! Use the -o option."),
		      Print::DEBUGGING, DebugPrint::INFO, "MMS");
    init_error++;
  }

  if (init_error)
    clean_up(0);

#ifdef mms_debug
  cerr << "At least one input and one output plugin were loaded... fine" << endl;
#endif


#ifdef mms_debug
  cerr << "Detecting specified output device" << endl;
#endif

  // Try to detect specified plugin (config or overwriten by command line)
  for (vector<OutputPlugin*>::iterator iter=plugins->op_data.begin(); iter!=plugins->op_data.end(); iter++)
    if ((*iter)->plugin_name() == conf->p_outdev()) {
#ifdef mms_debug
      cerr << "Specified device found: " << (*iter)->plugin_name();
#endif
      (*iter)->device->init();
      render->device=(*iter)->device;
      (*iter)->was_loaded_ok();
#ifdef mms_debug
      cerr << " and Init ok" << endl;
#endif
      break;
    }

  //  if no one found report and fail
  if (!render->device) {
    if (plugins->op_data.size() > 0) {
      OutputPlugin *out_dev = plugins->op_data.front();
#ifdef mms_debug
      cerr << "Not render device found, using first: " << 
	out_dev->plugin_name() << endl;
#endif
      out_dev->device->init();
      conf->s_outdev(out_dev->plugin_name());
      render->device=out_dev->device;
      out_dev->was_loaded_ok();
    } else {
      init_error++;
      cerr << "No output plugin can init" << endl;
    }
  }

#ifdef mms_debug
  cerr << "Detecting specified input device" << endl;
#endif

  InputMaster *im = S_InputMaster::get_instance();

  for (vector<InputPlugin*>::const_iterator iter=plugins->ip_data.begin();
       iter!=plugins->ip_data.end(); iter++)
    if ((*iter)->plugin_name() == conf->p_input()) {
#ifdef mms_debug
      cerr << "Specified device found: ";
#endif
      if ((*iter)->device->init()) {
        im->device=(*iter)->device;
#ifdef mms_debug
        cerr << (*iter)->plugin_name() << " and Init ok" << endl;
#endif
      } else
        cerr << "Device " << (*iter)->plugin_name() << " does not init" << endl;
      break;
    }

  // else use first one (but the one already tested!) that init
  if (!im->device) {
    cerr << "Input plugin not found - Trying first one" << endl;
    for (vector<InputPlugin*>::const_iterator iter=plugins->ip_data.begin();
         iter!=plugins->ip_data.end(); iter++)
      if (((*iter)->plugin_name() != conf->p_input()))
	{
#ifdef mms_debug
        cerr << "Device found " << (*iter)->plugin_name();
#endif
        if ((*iter)->device->init()) {
#ifdef mms_debug
          cerr << " and Init ok" << endl;
#endif
	  conf->s_input((*iter)->plugin_name());
          im->device=(*iter)->device;
          cerr << " Input device ok, using " << (*iter)->plugin_name() << endl;
          break;
        }
#ifdef mms_debug
	  else
	    cerr << " but does not init..." << endl;
#endif
      }
  }

  if (!im->device) {
    // FIXME
    cerr << "No input plugin can init" << endl;
    init_error++;
  }

  if (init_error)
    return 1;

#ifdef mms_debug
  cerr << "Input and Output plugin init OK" << endl;
#endif

#ifdef mms_debug
  cerr << "Init input master device" << endl;
#endif
  im->device->start();
#ifdef mms_debug
  cerr << "Init render device" << endl;
#endif
  render->device->start();

  Global *global = S_Global::get_instance();

  if (global->lcd_output_possible()) {
    global->lcd_add_output("  My Media System  ");
    global->lcd_print();
  }
  
#ifdef mms_debug
  cerr << "Start mms Startmenu mainloop" << endl;
#endif

  start();

  S_BackgroundUpdater::get_instance()->deactivate();
  S_ScreenUpdater::get_instance()->deactivate();

  S_Global::get_instance()->clean_up();

  deinit_mms(false);

  exit(0);
}
