#include "notify.hpp"

extern "C" {
#include <inotifytools/inotifytools.h>
#include <inotifytools/inotify.h>
}

#include "updater.hpp"

#include <cassert>

int Notify::notify_timeout()
{
  return 1000; // ms
}

void Notify::check_for_changes()
{
  // get plugins in a temporary structures to minimize locking of the
  // needs_update structure, since calling the callback function might take
  // some time
  std::set<notify_item> plugins_needing_update = get_plugins();

  for (std::set<notify_item>::const_iterator iter = plugins_needing_update.begin(), 
	 iend = plugins_needing_update.end(); iter != iend; ++iter)
    iter->func->call();
}

Notify::Notify()
{
  BackgroundUpdater *bg_updater = S_BackgroundUpdater::get_instance();

  TimeFunctori<Notify> *time_func = new TimeFunctori<Notify>(this, &Notify::notify_timeout);
  TimeFunctorv<Notify> *call_func = new TimeFunctorv<Notify>(this, &Notify::check_for_changes);
  bg_updater->timer.add(TimeElement("notify", time_func, call_func));
  
  inotifytools_initialize();
}

Notify::~Notify()
{
  for (std::vector<notify_item>::iterator iter = dirs_watching.begin(), iter_end = dirs_watching.end(); 
       iter != iter_end; ++iter) {

    delete iter->func;

    for (std::list<std::string>::const_iterator i = iter->dirs.begin(), iend = iter->dirs.end();
	 i != iend; ++i)
      inotifytools_remove_watch_by_filename(i->c_str());
  }
}

// they should be notified back somehow?
void Notify::register_plugin(const std::string& name, const std::list<std::string>& dir_passed, TimeFv *func)
{
  lock.enterMutex();

  std::list<std::string> dirs;

  for (std::list<std::string>::const_iterator i = dir_passed.begin(), iend = dir_passed.end();
       i != iend; ++i) {

    std::string dir = *i;

    // dirs must end with "/"
    if (dir[dir.size()-1] != '/')
      dir += '/';

    int status = inotifytools_watch_file(dir.c_str(), IN_CREATE | IN_DELETE | IN_DELETE_SELF | IN_MODIFY | IN_MOVE_SELF | IN_MOVE);

    if (status == 0) {
      delete func;
      lock.leaveMutex();
      return;
    }

    dirs.push_back(dir);
  }
  
  bool already_registered = false;
  
  for (std::vector<notify_item>::iterator iter = dirs_watching.begin(), iter_end = dirs_watching.end(); 
       iter != iter_end; ++iter) {
    if (iter->plugin_name == name) {

      for (std::list<std::string>::const_iterator i = iter->dirs.begin(), iend = iter->dirs.end();
	   i != iend; ++i)
	inotifytools_remove_watch_by_filename(i->c_str());

      // sanity check, for double add
      assert(iter->dirs != dirs);

      iter->dirs = dirs;

      delete iter->func;
      iter->func = func;
      
      already_registered = true;
      break;
    }
  }
  
  if (!already_registered) {
    struct notify_item item;
    item.plugin_name = name;
    item.dirs = dirs;
    item.func = func;
    dirs_watching.push_back(item);
  }

  lock.leaveMutex();
}

void Notify::run()
{
  struct inotify_event *event;
  char filename[4096];

  while (true){
    event = inotifytools_next_event(-1);
    inotifytools_snprintf(filename, 4096, event, "%w");

    lock.enterMutex();

    bool found_match = false;

    for (std::vector<notify_item>::iterator iter = dirs_watching.begin(), iter_end = dirs_watching.end(); 
	 iter != iter_end; ++iter) {
      for (std::list<std::string>::iterator i = iter->dirs.begin(), iend = iter->dirs.end(); 
	   i != iend; ++i) {
	if (filename == *i) {
	  if (needs_update.find(*iter) == needs_update.end())
	    needs_update.insert(*iter);
	  std::cout << needs_update.size() << std::endl;
	  found_match = true;
	  break;
	}
      }
      if (found_match)
	break;
    }
    lock.leaveMutex();
  }
}

const std::set<notify_item> Notify::get_plugins()
{
  lock.enterMutex();
  std::set<notify_item> plugins = needs_update;
  needs_update.clear();
  lock.leaveMutex();
  return plugins;
}
