// used for #ifdef
#include "config.h"

#include "libfspp.hpp"

#include "audio.hpp"

#include "common.hpp"
#include "graphics.hpp"
#include "print.hpp"
#include "plugins.hpp"
#include "gettext.hpp"

#include "updater.hpp"
#include "extra_menu.hpp"

#include "input.hpp"
#include "printer.hpp"

#ifdef use_inotify
#include "notify.hpp"
#endif

// cd 
#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "kernelheaders.hpp"

// time 
#include <time.h>

// sleep
#include <unistd.h>

using std::list;
using std::vector;
using std::string;

Audio::Audio() 
  : exit_loop(false), change_mode(false), 
    printing_information(false), 
#ifdef use_media
    mounted_media(false),
#endif
    playlist_id(0), 
    was_playing(false), was_started(false), pos(0), 
    radio_stations_downloaded(false), 
    disabled_busy_indicator(false), display_message(true)
{
  input_master->parse_keys("playlist", "1");
  input_master->parse_keys("audio", "4");
  input_master->extend_default("audio", "2");

  audio_conf = S_AudioConfig::get_instance();

  mode = ADD;

  audio_state = S_Audio_s::get_instance();
  audio_state->set_audio(this);
  opts.set_audio(this);

  add_playback_keys();
}

void Audio::save_runtime_settings()
{
  std::ofstream file;

  string path = conf->p_var_data_dir() + "options/Audio Runtime";

  file.open(path.c_str());
    
  if (!file) {
    DebugPrint perror(gettext("Could not write options to file ") + path,
		      Print::DEBUGGING, DebugPrint::CRITICAL, "AUDIO");
  } else {
    file << "volume," << audio_state->p->getvol() << std::endl;
    file << "playlist_pos," << get_real_playlist_pos() << std::endl;
    file << "play_now_warning," << play_now_warning << std::endl;
  }

  file.close();
}

void Audio::begin_external()
{
  if (!was_started) { // protect against calling this function twice
    was_playing = false;
    was_started = false;
    pos = 0;
  }

  if (audio_conf->p_audio_started() && audio_state->p->loaded()) {
    was_started = true;

    if (audio_state->p->is_playing()) {
      was_playing = true;
      pos = audio_state->p->getpos();
      track = audio_state->p->p_cur_nr();
      audio_state->p->stop();
      sleep(1);
    }

    // stop the updater as it otherwise might query the player
    ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

    if (screen_updater->timer.status("audio_next"))
      screen_updater->timer.deactivate("audio_next");
    if (screen_updater->timer.status("audio_update"))
      screen_updater->timer.deactivate("audio_update");
    if (screen_updater->timer.status("audio_fullscreen"))
      screen_updater->timer.deactivate("audio_fullscreen");
    if (global->lcd_output_possible() && screen_updater->timer.status("audio_lcd"))
      screen_updater->timer.deactivate("audio_lcd");

    audio_state->p->release_device();
  }
}

void Audio::end_external()
{
  audio_state->p->restore_device();

  if (was_started) {

    was_started = false;

    if (was_playing) {
#ifdef use_media
      audio->check_mount_before(track);
#endif
      audio_state->p->addfile(track);
      audio_state->p->set_cur_nr(track);
      audio_state->p->setpos(pos);

      // restore volume
      audio_state->p->setvol(audio_state->p->getvol());
    }

    // restart the updater
    ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();
    if (!screen_updater->timer.status("audio_next"))
      screen_updater->timer.activate("audio_next");
    if (!screen_updater->timer.status("audio_update"))
      screen_updater->timer.activate("audio_update");
    if (!screen_updater->timer.status("audio_fullscreen"))
      screen_updater->timer.activate("audio_fullscreen");
    if (global->lcd_output_possible() && !screen_updater->timer.status("audio_lcd"))
      screen_updater->timer.activate("audio_lcd");
  }
}

#ifdef use_media
void Audio::pos_umount()
{
  if (mounted_media) {
    run::external_program("umount " + cd->get_mount_point());
    mounted_media = false;
  }
}

void Audio::check_mount_before(const audio_type& track)
{
  if (!mounted_media && track.type == "media-file") {
    run::external_program("mount " + cd->get_mount_point());
    mounted_media = true;
  } 
}

void Audio::check_mount_after(const audio_type& track)
{
  if (mounted_media && track.type != "media-file") {
    run::external_program("umount " + cd->get_mount_point());
    mounted_media = false;
  }
}
#endif

bool Audio::file_sort::operator()(const Simplefile& lhs, const Simplefile& rhs)
{
  return sort_order(lhs.lowercase_name, rhs.lowercase_name, lhs.type == "dir", rhs.type == "dir");
}

// commands
void Audio::play_now() 
{
  // FIXME:
#if 0
  if (add_dir(vector_lookup(*files,position_int()).path).size() == 0) {
    DialogWaitPrint pdialog(gettext("Folder is empty"), 1000);
    return;
  }
  
  if (play_now_warning) {
    DialogWaitPrint pdialog(5000);
    pdialog.add_line(gettext("Clearing playlist to play contents of folder"));
    pdialog.add_line(gettext("The old playlist has been saved as 'autosaved'"));
    pdialog.print();
    play_now_warning = false;
  }

  save_playlist("autosaved", true);

  clear_playlist();
  // don't change mode to playlist
  change_mode = false;
  display_message = false;
  add();
  display_message = true;
  play_track();
#endif
}

void Audio::play_track_now() 
{
  // FIXME:
#if 0
  add();
  playlist_pos_int = playlist.size()-1;
  play_track();
#endif
}

void Audio::exit()
{
  exit_loop = true;
}

void Audio::leave_dir()
{
  if (folders.size() != 1) {
    folders.pop();

    load_current_dirs();

#ifdef use_inotify
    S_Notify::get_instance()->register_plugin("audio", folders.top().first,
					      new TimeFunctorv<Audio>(this, &Audio::reparse_current_dir));
#endif
  }
}

class SaveFileFunctor
{
 private:
  void (*fpt)(const string&, int, Overlay&);
  Overlay o;
  
 public:

  SaveFileFunctor(void(*_fpt)(const string&, int, Overlay&))
    : o(Overlay("fixed size dialog"))
  { fpt=_fpt; };

  ~SaveFileFunctor()
  {
    Render *render = S_Render::get_instance();
    // fixme: wait and aquire as to wait?
    render->image_mut.enterMutex();
    o.cleanup();
    render->image_mut.leaveMutex();
  }

  void call(const string& r)
  { return (*fpt)(gettext("Enter filename: ") + r, 450, o); };
};

void Audio::save_playlist_func()
{
  BusyIndicator::get_instance().idle();

  string name = "";
  list<Input> exit_inputs;
  Input i;
  i.command = "back"; i.mode = "general";
  exit_inputs.push_back(i);
  SaveFileFunctor *print_func = new SaveFileFunctor(&print_dialog_fixed_size);
  std::pair<bool, Input> res = input_master->device->generate_string(name, print_func, exit_inputs);
  delete print_func;
  usleep(100*1000);
  if (res.first)
    save_playlist(name);
}

string Audio::format_time(int cur_time, int total_time)
{
  char buffer [512];
  snprintf(buffer, 512, "%02d:%02d / %02d:%02d", seconds_to_minutes(cur_time), 
	   remaining_seconds(cur_time), seconds_to_minutes(total_time), remaining_seconds(total_time));
  return buffer;
}

string Audio::format_time(int time)
{
  char buffer [512];
  snprintf(buffer, 512, "%02d:%02d", seconds_to_minutes(time), remaining_seconds(time));
  return buffer;
}

const int Audio::position_int()
{
  if (mode == ADD)
    return folders.top().second;
  else
    return playlist_pos_int;
}

void Audio::set_position_int(int new_pos)
{
  if (mode == ADD)
    folders.top().second = new_pos;
  else
    playlist_pos_int = new_pos;
}

bool Audio::choose_playlist()
{
  BusyIndicator &busy_indicator = BusyIndicator::get_instance();

  input_master->save_map();
  input_master->set_map("playlist");

  Shutdown *sd = S_Shutdown::get_instance();

  Input input;
  
  saved_playlists = list_playlists();
 
  saved_playlist_pos = 0;

  returned = false;
  return_value = false;

  bool update_needed = true;

  while (!(input.command == "back" && input.mode == "general"))
    {
      if (update_needed)
	print_list(saved_playlists, vector_lookup(saved_playlists, pos), pos, true);
      
      busy_indicator.idle();

      input = input_master->device->get_input();

      busy_indicator.busy();
      
      if (sd->is_enabled()) {
      	sd->cancel();
      	continue;
      }

      update_needed = true;

      if (input.command == "prev")
 	{
 	  if (pos != 0)
	    --pos;
	  else 
	    pos = saved_playlists.size()-1;
 	}
      else if (input.command == "next")
 	{
 	  if (pos != saved_playlists.size()-1)
	    ++pos;
	  else
	    pos = 0;
 	}
      else if (input.command == "page_up")
 	{
	  if (saved_playlists.size() > conf->p_jump()) {
	    int new_pos = saved_playlist_pos-conf->p_jump();
	    if (saved_playlist_pos == 0)
	      saved_playlist_pos = saved_playlists.size()-1+new_pos;
	    else if (new_pos < 0)
	      saved_playlist_pos = 0;
	    else
	      saved_playlist_pos = new_pos;
	  }
 	}
      else if (input.command == "page_down")
 	{
	  if (saved_playlists.size() > conf->p_jump()) {
	    if (saved_playlist_pos > (saved_playlists.size() - conf->p_jump()) && saved_playlist_pos != (saved_playlists.size()-1))
	      saved_playlist_pos = saved_playlists.size()-1;
	    else 
	      saved_playlist_pos = (saved_playlist_pos+conf->p_jump())%saved_playlists.size();
	  }
 	}
      else if (input.command == "delete")
	{
	  delete_playlist();
	  
	  if (returned)
	    return return_value;
	}
      else if (input.command == "action")
	{
	  action_load_playlist();
	  
	  if (returned)
	    return return_value;
	}
      else if (input.command == "playlist_mode" && input.mode == "audio_general")
	{
	  return false;
	}
      else	
	update_needed = !global->check_commands(input);
    }

  input_master->restore_map();

  return false;
}

void Audio::action_load_playlist()
{
  if (load_playlist(vector_lookup(saved_playlists, saved_playlist_pos))) {
    audio_state->p->stop();
    returned = true;
    return_value = true;
  } else
    DialogWaitPrint pdialog(gettext("Playlist is corrupt"), 1000);
}

void Audio::delete_playlist()
{
  // delete the file from drive
  std::ostringstream del;
  del << "rm \"" + conf->p_var_data_dir() + "playlists/" << vector_lookup(saved_playlists, pos) << "\"" << std::endl;
  run::external_program(del.str());
         
  vector<string>::iterator i = saved_playlists.begin() + saved_playlist_pos;
  saved_playlists.erase(i);

  // this is saved_playlists.size(), not -1, because we just deleted a track
  if (saved_playlist_pos == saved_playlists.size())
    --saved_playlist_pos;

  // if we removed the last element then bail
  if (saved_playlists.size() == 0) {
    returned = true;
    return_value = false;
  }
}

void Audio::choose_radio()
{
  // init it if it's not already done
  if (!audio_conf->p_audio_started())
    audio_state = S_Audio_s::get_instance();

  Input input;

  vector<std::pair<string, string> > cur_files;
  vector<string> print_files;

  std::pair<string, string> my_radio_genre("My Radio Stations", "");
  print_files.push_back(my_radio_genre.first);
  cur_files.push_back(my_radio_genre);

  Shoutcast *shoutcast = Shoutcast::get_instance();

  if (audio_conf->p_download_radio()) {
    if (!radio_stations_downloaded) {
      DialogWaitPrint pdialog(gettext("Downloading radio stations"), 2000);
      radio_stations_downloaded = true;
    }

    list<std::pair<string, string> > shoutcast_genrelist = shoutcast->get_genrelist();  

    for (list<std::pair<string, string> >::iterator i = shoutcast_genrelist.begin(),
	   end = shoutcast_genrelist.end(); i != end; ++i) {
      print_files.push_back(i->first);
      cur_files.push_back(*i);
    }
  }

  radio_pos = 0;
  old_radio_pos = 0;

  radio_update_needed = true;
  bool exit_radio = false;
  in_genre = true;

  BusyIndicator &busy_indicator = BusyIndicator::get_instance();

  while (!(input.command == "back" && input.mode == "general" && exit_radio))
    {
      if (radio_update_needed)
	print_list(print_files, vector_lookup(cur_files, radio_pos).first, radio_pos, false);

      busy_indicator.idle();

      input = input_master->device->get_input();

      busy_indicator.busy();

      radio_update_needed = true;

      if (input.command == "prev")
 	{
 	  if (radio_pos != 0)
	    --radio_pos;
	  else
	    radio_pos = cur_files.size()-1;
 	}
      else if (input.command == "next")
 	{
 	  if (radio_pos != cur_files.size()-1)
	    ++radio_pos;
	  else
	    radio_pos = 0;
 	}
      else if (input.command == "page_up")
 	{
	  if (cur_files.size() > conf->p_jump()) {
	    int new_radio_pos = radio_pos-conf->p_jump();
	    if (radio_pos == 0)
	      radio_pos = cur_files.size()-1+new_radio_pos;
	    else if (new_radio_pos < 0)
	      radio_pos = 0;
	    else
	      radio_pos = new_radio_pos;
	  }
 	}
      else if (input.command == "page_down")
 	{
	  if (cur_files.size() > conf->p_jump()) {
	    if (radio_pos > (cur_files.size() - conf->p_jump()) && radio_pos != (cur_files.size()-1))
	      radio_pos = cur_files.size()-1;
	    else
	      radio_pos = (radio_pos+conf->p_jump())%cur_files.size();
	  }
	}
      // play specific
      else if (input.command == "action")
	{
	  add_radio_track(cur_files, print_files);
        }
      else if(input.command == "back")
	{
	  if(in_genre)
            exit_radio = true;
	  else
	    {
	      print_files.clear();
	      cur_files.clear();

	      print_files.push_back(my_radio_genre.first);
	      cur_files.push_back(my_radio_genre);
	    
	      if (audio_conf->p_download_radio()) {
		list<std::pair<string, string> > shoutcast_genrelist = shoutcast->get_genrelist();

		for (list<std::pair<string, string> >::iterator i = shoutcast_genrelist.begin(),
		       end = shoutcast_genrelist.end(); i != end; ++i) {
		  print_files.push_back(i->first);
		  cur_files.push_back(*i);
		}
	      }

	      radio_pos = old_radio_pos;
	      radio_update_needed = true;
	      in_genre = true;
	    }
	}
      else
	radio_update_needed = !global->check_commands(input);
    }
}

void Audio::print_string_element(const string& r, const string& position, int y)
{
  string name = r;
  string_format::format_to_size(name, "Vera/16", conf->p_v_res()-160, true);

  if (r == position) {
    int width = conf->p_v_res()-(2*67);
    if (conf->p_v_res() == 720)
      width -= 5;

    render->current.add(new PObj(themes->audio_marked_left, 67, y-2, 2, NOSCALING));
    render->current.add(new PObj(themes->audio_marked, 67+25, y-2, width-50, 2));
    render->current.add(new PObj(themes->audio_marked_right, 67+25+width-50, y-2, 2, NOSCALING));
  }

  if (conf->p_convert())
    name = string_format::convert(name);

  render->current.add(new TObj(name, "Vera/16", 75, y, 
			       themes->audio_font1, themes->audio_font2, 
			       themes->audio_font3, 3));
}

void Audio::print_top()
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->audio_background, 0, 0, 0, SCALE_FULL));

  render->current.add(new PObj(themes->startmenu_music_dir, 25, 10, 2, NOSCALING));

  string header = gettext("Audio - Add files");
  if (mode == PLAY)
    header = gettext("Audio - Playlist");

  if (themes->show_header) {
    render->current.add(new TObj(header, "Vera/28", 100, 10, 
				 themes->audio_header_font1, themes->audio_header_font2, 
				 themes->audio_header_font3, 2));
  }

  if (search_mode) {

    int y = 60;
    if (conf->p_v_res() == 720) {
      y -= 3;
    }

    render->current.add(new RObj(0, y, conf->p_v_res(), 45, 
				 themes->general_search_rectangle_color1, 
				 themes->general_search_rectangle_color2, 
				 themes->general_search_rectangle_color3, 
				 themes->general_search_rectangle_color4, 2));

    render->current.add(new PObj(themes->general_search, 47, y, 3, NOSCALING));

    render->current.add(new TObj(gettext("Search: ") + search_str, "Vera/22", 100, y+5, 
				 themes->audio_font1, themes->audio_font2, 
				 themes->audio_font3, 3));
  }
}

void Audio::print_buttom(int cur_files_size)
{
  std::ostringstream pos_in_list;

  if (search_mode)
    if (cur_files_size > 0)
      pos_in_list << offset + 1 << "/" << cur_files_size;
    else
      pos_in_list << "";
  else
    pos_in_list << position_int()+1 << "/" << cur_files_size;
  
  int x = string_format::calculate_string_width(pos_in_list.str(), "Vera/18");
  
  render->current.add(new TObj(pos_in_list.str(), "Vera/18", 
			       conf->p_v_res()-(60+x), 20,
			       themes->audio_font1, themes->audio_font2, 
			       themes->audio_font3, 3));

  int opts_pos=46;

  // check if repeat is on then display
  if (opts.repeat()== gettext("yes") && opts.shutdown()== gettext("no")) {
    render->current.add(new PObj(themes->audio_repeat, conf->p_v_res()-opts_pos ,conf->p_h_res()-52, 3, NOSCALING));
    opts_pos = opts_pos + 50; 
  }
  
  // check if shutdown is on then display
  if (opts.shutdown()== gettext("yes")) {
    render->current.add(new PObj(themes->audio_shutdown, conf->p_v_res()-opts_pos ,conf->p_h_res()-52, 3, NOSCALING));
    opts_pos = opts_pos + 50; 
  }
  
  // display shuffle mode
  if (opts.shuffle()== gettext("real random")) {
    render->current.add(new PObj(themes->audio_random, conf->p_v_res()-opts_pos, conf->p_h_res()-52, 3, NOSCALING));
    opts_pos = opts_pos + 50;
  }
  else if (opts.shuffle()== gettext("intelligent")) {
    render->current.add(new PObj(themes->audio_randomi, conf->p_v_res()-opts_pos, conf->p_h_res()-52, 3, NOSCALING));
    opts_pos = opts_pos + 50; 
  }
  else if (opts.shuffle()== gettext("winamp random")) {
    render->current.add(new PObj(themes->audio_randomw,  conf->p_v_res()-opts_pos, conf->p_h_res()-52, 3, NOSCALING)); 
    opts_pos = opts_pos + 50;
  }

  render->draw_and_release("Audio");
}

void Audio::print_list(const vector<string>& cur_files, const string& position, const int int_position, bool is_playlist)
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->audio_background, 0, 0, 0, SCALE_FULL));
  
  if (is_playlist) {
    render->current.add(new PObj(themes->audio_playlist_icon, 25, 10, 2, NOSCALING));
    render->current.add(new TObj(gettext("Audio - Choose Playlist"), "Vera/28", 100, 10, 
				 themes->audio_header_font1, themes->audio_header_font2, 
				 themes->audio_header_font3, 2));
  } else {
    render->current.add(new PObj(themes->startmenu_radio, 25, 10, 2, NOSCALING));
    render->current.add(new TObj(gettext("Audio - Radio"), "Vera/28", 100, 10, 
				 themes->audio_header_font1, themes->audio_header_font2, 
				 themes->audio_header_font3, 2));
  }

  PrintFunctor<Audio, string> print_func(this, &Audio::print_string_element);
  print_range(cur_files, position, int_position, &print_func);

  if (is_playlist)
    render->draw_and_release("Audio choose playlist");
  else
    render->draw_and_release("Audio radio");
}

void Audio::detect_cdformat() 
{
  int cdrom = cd->cdrom;

  switch(ioctl(cdrom, CDROM_DISC_STATUS)) {
    
  case CDS_NO_INFO:
  case CDS_NO_DISC:
  case CDS_TRAY_OPEN:
  case CDS_DRIVE_NOT_READY:
  case CDS_DISC_OK:
    {
      Print pdialog(gettext("The disc does not contain any recognizable files"), 
		    Print::SCREEN);
    }
    break;
  case CDS_AUDIO:
    cd_menu();
    break;
  case CDS_DATA_1:
  case CDS_DATA_2:
  case CDS_XA_2_1:
  case CDS_XA_2_2:
    add_data_cd();
    break;
  default:
    {
      Print pdialog(gettext("The disc does not contain any recognizable files"), 
		    Print::SCREEN);
    }
    break;
  }
}

vector<string> Audio::list_playlists() 
{
  vector<string> cur_files;

  string argv = conf->p_var_data_dir() + "playlists/";
  
  for (file_iterator<file_t, default_order> i (argv); (i != i.end ()); i.advance(false))
    {
      string filename = i->getName();
      cur_files.push_back(filename.substr(argv.size()+1));
    }

  return cur_files;
}

void Audio::print_audiotrack()
{
  if (audio_conf->p_audio_started()) {

    if (audio_state->p_playing() && audio_state->p->is_playing() 
	&& !audio_state->fullscreen_info && !audio_state->p->is_buffering()) {

      render->wait_and_aquire();

      audio_state->has_played = true;

      if (audio_state->playback.elements.size() > 0)
	audio_state->playback.partial_cleanup();

      int x_pos = 7;
      string buffer, artist, album, title;
      
      print_audiotrack_helper(x_pos, buffer, artist, album, title);

      if (audio_state->p_pause()) {
	PObj *p = new PObj(themes->audio_playback_pause, x_pos, conf->p_h_res()-65, 0, NOSCALING);
	PObj *p2 = new PObj(themes->audio_playback_pause, x_pos + (53 - p->real_w)/2, 
			    conf->p_h_res() - 65 + (65 - p->real_h)/2, 0, NOSCALING);
	delete p;
	audio_state->playback.add(p2);
      } else {
	PObj *p = new PObj(themes->audio_playback_play, x_pos, conf->p_h_res()-65, 0, NOSCALING);
	PObj *p2 = new PObj(themes->audio_playback_play, x_pos + (53 - p->real_w)/2, 
			    conf->p_h_res() - 65 + (65 - p->real_h)/2, 0, NOSCALING);
	delete p;
	audio_state->playback.add(p2);
      }
      
      // print artist info
      if ((!artist.empty() && !title.empty()) || 
	  (audio_state->p->p_cur_nr().type == "web" && !title.empty())) {

	if (artist == "SCRIPT") {
	  artist = "";
	  album = "";
	}

	if (artist.empty() && audio_state->p->p_cur_nr().type == "web")
	  artist = audio_state->p->p_cur_nr().name; // radio station name

	int y_delta = 0;
	
	if (artist.empty() && album.empty())
	  y_delta = - 15;
	else {

	  int artist_x_size = string_format::calculate_string_width(artist, "Vera/17");
	  int album_x_size = string_format::calculate_string_width(album, "Vera/17");

	  int max_size = (conf->p_v_res()-(53+x_pos)-30-146);
	  int max_artist_size = max_size/2;
	  int max_album_size = max_size/2;

	  if (artist_x_size < max_size/2)
	    max_album_size = max_size-artist_x_size;
	  else
	    if (album_x_size < max_size/2)
	      max_artist_size = max_size-album_x_size;

	  string_format::format_to_size(artist, "Vera/17", max_artist_size, true);
	  string_format::format_to_size(album, "Vera/17", max_album_size, true);

	  string artist_album = artist;
	  if (artist.empty() || album.empty())
	    artist_album += album;
	  else
	    artist_album += (" - " + album);

	  audio_state->playback.add(new TObj(artist_album, 
					     "Vera/17", 53+x_pos, conf->p_h_res()-65, 
					     themes->audio_playback_font1, themes->audio_playback_font2, 
					     themes->audio_playback_font3, 0));
	
	}

	int x_size = string_format::calculate_string_width(string("     ") + buffer, "Vera/17");

	string_format::format_to_size(title, "Vera/17", conf->p_v_res()-(53+x_pos)-x_size-146, true);

	audio_state->playback.add(new TObj(title + "     " + buffer, 
					   "Vera/17", 53+x_pos, conf->p_h_res()-35+y_delta, 
					   themes->audio_playback_font1, themes->audio_playback_font2, 
					   themes->audio_playback_font3, 0));

    
      } else {
	Simplefile cur_pos = audio_state->p->p_cur_nr();

	string name = cur_pos.name  + "     " + buffer;

	string_format::format_to_size(name, "Vera/17", conf->p_v_res()-2*(53+x_pos), true);
	audio_state->playback.add(new TObj(name, "Vera/17", 53+x_pos, conf->p_h_res()-51, 
					   themes->audio_playback_font1, themes->audio_playback_font2,
					   themes->audio_playback_font3, 0));

      }

      render->draw_and_release("Audio print track", true);

    } else if (audio_state->has_played) {

      render->wait_and_aquire();

      audio_state->has_played = false;
      audio_state->playback.cleanup();

      render->image_mut.leaveMutex();
    }
  }
}

void Audio::print_lcd_audiotrack()
{
  if (audio_conf->p_audio_started()) {
    
    if (audio_state->p_playing() && audio_state->p->is_playing() && 
	time(0)-conf->p_last_key() > 10) {

      string buffer = format_time(audio_state->p->p_cur_time(), audio_state->p->p_total_time());
  
      if (!audio_state->p->p_artist().empty() && !audio_state->p->p_title().empty()) {
	global->lcd_add_output(string_format::pretty_print_string_copy(audio_state->p->p_artist(), 20));
	global->lcd_add_output(string_format::pretty_print_string_copy(audio_state->p->p_album(), 20));
	global->lcd_add_output(string_format::pretty_print_string_copy(audio_state->p->p_title(), 20));
	global->lcd_add_output(string(buffer));
	global->lcd_print();
      } else {
	Simplefile cur_pos = audio_state->p->p_cur_nr();

	global->lcd_add_output(string_format::pretty_print_string_copy(cur_pos.name, 20));
	global->lcd_add_output("");
	global->lcd_add_output("");
	global->lcd_add_output(string(buffer));
	global->lcd_print();
      }
    }
  }
}

int Audio::print_audio_time()
{
  return 300; // ms
}

int Audio::check_audio_time()
{
  return 100; // ms
}

void Audio::print_audio_screensaver()
{
  if (audio_state->p_playing() && audio_state->p->is_playing()) {

    render->wait_and_aquire();
  
    if (audio_state->playback_fullscreen.elements.size() > 0)
      audio_state->playback_fullscreen.partial_cleanup();

    audio_state->playback_fullscreen.add(new RObj(0, 0, conf->p_v_res(), conf->p_h_res(), 0, 0, 0, 255, 0));

    int tmp;
    string buffer, artist, album, title;
      
    print_audiotrack_helper(tmp, buffer, artist, album, title);

    int max_x_pic = 240, max_y_pic = 266;
	  
    if (conf->p_v_res() == 1024) {
      max_x_pic = 300;
      max_y_pic = 333;
    } else if (conf->p_h_res() == 405) {
      max_x_pic = 135;
      max_y_pic = 150;
    }
    
    int x_pos = max_x_pic+60+30;
    int y = 0;

    if ((!artist.empty() && !title.empty())
	|| (audio_state->p->p_cur_nr().type == "web" && !title.empty())) {

      y = (conf->p_h_res()-165)/2;
      
      print_audio_screensaver_helper(x_pos, y, max_x_pic, max_y_pic);
      
      if (artist == "SCRIPT") {
	artist = "";
	album = "";
      }

      if (artist.empty() && audio_state->p->p_cur_nr().type == "web")
	artist = audio_state->p->p_cur_nr().name; // radio station name

      string_format::format_to_size(artist, "Vera/28", conf->p_v_res()-x_pos-30, true);
      audio_state->playback_fullscreen.add(new TObj(artist, "Vera/28", x_pos, y, 
						    themes->audio_playback_font1, themes->audio_playback_font2, 
						    themes->audio_playback_font3, 1));

      y += 45;

      string_format::format_to_size(album, "Vera/25", conf->p_v_res()-x_pos-30, true);
      audio_state->playback_fullscreen.add(new TObj(album, "Vera/25", x_pos, y, 
						    themes->audio_playback_font1, themes->audio_playback_font2, 
						    themes->audio_playback_font3, 1));

      y += 40;

      string_format::format_to_size(title, "Vera/22", conf->p_v_res()-x_pos-30, true);
      audio_state->playback_fullscreen.add(new TObj(title, "Vera/22", x_pos, y, 
						    themes->audio_playback_font1, themes->audio_playback_font2, 
						    themes->audio_playback_font3, 1));

      y += 40;

      audio_state->playback_fullscreen.add(new TObj(buffer, "Vera/22", x_pos, y, 
						    themes->audio_playback_font1, themes->audio_playback_font2, 
						    themes->audio_playback_font3, 1));

      y += 40;

    } else {

      print_audio_screensaver_helper(x_pos, tmp, max_x_pic, max_y_pic);

      y = (conf->p_h_res()-80)/2;

      Simplefile cur_pos = audio_state->p->p_cur_nr();

      string name = cur_pos.name;

      string_format::format_to_size(name, "Vera/28", conf->p_v_res()-x_pos-30, true);
      audio_state->playback_fullscreen.add(new TObj(name, "Vera/22", x_pos, y, 
						    themes->audio_playback_font1, themes->audio_playback_font2,
						    themes->audio_playback_font3, 1));

      y += 40;
      
      audio_state->playback_fullscreen.add(new TObj(buffer, "Vera/22", x_pos, y, 
						    themes->audio_playback_font1, themes->audio_playback_font2,
						    themes->audio_playback_font3, 1));
      y += 40;

    }

    y += 20;

    std::ostringstream played_info;
    played_info << gettext("Playing track: ");
    
    if (opts.shuffle() != gettext("off"))
      played_info << audio_state->played_tracks.size();
    else
      played_info << get_real_playlist_pos()+1;

    played_info << "/" << playlist_size();

    audio_state->playback_fullscreen.add(new TObj(played_info.str(), "Vera/22", x_pos, y, 
						  themes->audio_playback_font1, themes->audio_playback_font2, 
						  themes->audio_playback_font3, 1));

    y += 40;

    if (audio_state->p_pause())
      audio_state->playback_fullscreen.add(new PObj(themes->audio_playback_pause, x_pos, y, 0, NOSCALING));

    render->draw_and_release("screensaver", true);
  }
}

int Audio::check_audio_fullscreen()
{
  return 300; // ms
}

void Audio::show_volume()
{
  render->wait_and_aquire();

  if (audio_state->volume.elements.size() > 0)
    audio_state->volume.partial_cleanup();

  if (audio_state->fullscreen_info) {

    if (audio_state->p->getvol()!=0)
      audio_state->volume.add(new TObj("VOL: " + conv::itos(audio_state->p->getvol()), "Vera/22", 
				       conf->p_v_res()-150, conf->p_h_res()-75, 
				       themes->audio_playback_font1, 
				       themes->audio_playback_font2, themes->audio_playback_font3, 1));

    if (audio_state->p->is_mute())
      audio_state->volume.add(new PObj(themes->audio_mute, conf->p_v_res()-200, conf->p_h_res()-75, 1, NOSCALING)); 
  } else {
    // display sound level 
    if (audio_state->p->getvol() > 0)
      audio_state->volume.add(new PObj(themes->audio_volume_bar, conf->p_v_res()-40, 
				       int(conf->p_h_res() - 190 + (115  * (1-audio_state->p->getvol()/100.0))), 
				       audio_state->p->getvol()/100.0, 3));

    // display skin for sound level
    audio_state->volume.add(new PObj(themes->audio_volume, conf->p_v_res()-40 ,conf->p_h_res()-190, 3, NOSCALING));

    // display muted
    if (audio_state->p->is_mute())
      audio_state->volume.add(new PObj(themes->audio_mute, conf->p_v_res()-246, conf->p_h_res()-52, 3, NOSCALING)); 
  }

  audio_state->volume_shown = true;

  displayed_volume = time(0);

  render->draw_and_release("volume", true);
}

int Audio::check_volume()
{
  return 300; // ms
}

void Audio::clean_up_volume()
{
  if (audio_state->volume_shown && time(0)-displayed_volume > 2) {

    render->wait_and_aquire();

    audio_state->volume_shown = false;
    audio_state->volume.cleanup();

    render->image_mut.leaveMutex();
  }
}

void Audio::set_audio_player()
{
  bool audio_player_set = false;
  Plugins *plugins = S_Plugins::get_instance();
  for (vector<AudioPlayerPlugin*>::const_iterator iter=plugins->audio_players.begin(), 
	 iend=plugins->audio_players.end(); iter!=iend; ++iter)
    if ((*iter)->plugin_name() == audio_conf->p_audio_player()) {
      audio_state->set_audio_player((*iter)->player);
      audio_player_set = true;
      break;
    }
  
  if (!audio_player_set) {
    DebugPrint perror(gettext("Could not find audio player plugin"), 
		      Print::DEBUGGING, DebugPrint::CRITICAL, "Audio");
    clean_up(0);
  }
}

void Audio::startup_updater()
{
  set_audio_player();

  load_last_playlist();

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();
  TimeFunctori<Audio> *time_func1 = new TimeFunctori<Audio>(this, &Audio::check_audio_time);
  TimeFunctorv<Audio> *call_func1 = new TimeFunctorv<Audio>(this, &Audio::check_audio_state);
  screen_updater->timer.add(TimeElement("audio_next", time_func1, call_func1));

  if (global->lcd_output_possible()) {
    TimeFunctori<Audio> *time_func2 = new TimeFunctori<Audio>(this, &Audio::print_audio_time);
    TimeFunctorv<Audio> *call_func2 = new TimeFunctorv<Audio>(this, &Audio::print_lcd_audiotrack);
    screen_updater->timer.add(TimeElement("audio_lcd", time_func2, call_func2));
  }

  TimeFunctori<Audio> *time_func3 = new TimeFunctori<Audio>(this, &Audio::print_audio_time);
  TimeFunctorv<Audio> *call_func3 = new TimeFunctorv<Audio>(this, &Audio::print_audiotrack);
  screen_updater->timer.add(TimeElement("audio_update", time_func3, call_func3));

  if (conf->p_idle_time() > 0) {
    TimeFunctori<Audio> *time_func4 = new TimeFunctori<Audio>(this, &Audio::check_audio_fullscreen);
    TimeFunctorv<Audio> *call_func4 = new TimeFunctorv<Audio>(this, &Audio::print_audio_fullscreen);
    screen_updater->timer.add(TimeElement("audio_fullscreen", time_func4, call_func4));
  }

  list<string> triggers;
  triggers.push_back("vol+");
  triggers.push_back("vol-");

  TimeFunctorv<Audio> *call_func5 = new TimeFunctorv<Audio>(this, &Audio::show_volume);
  screen_updater->trigger.add(TriggerElement("audio_volume", triggers, call_func5, 0));

  TimeFunctori<Audio> *time_func6 = new TimeFunctori<Audio>(this, &Audio::check_volume);
  TimeFunctorv<Audio> *call_func6 = new TimeFunctorv<Audio>(this, &Audio::clean_up_volume);
  screen_updater->timer.add(TimeElement("audio_volume", time_func6, call_func6));
}

void Audio::pause()
{
  audio_state->p->pause();
}

void Audio::stop()
{
  audio_state->p->stop();

  // if clear on finish = false
  if (playlist_size() == 0) {
    Simplefile empty;
    audio_state->p->set_cur_nr(empty);
  }
}

void Audio::next_track()
{
  audio_state->p->next();
}

void Audio::prev_track()
{
  audio_state->p->prev();
}

void Audio::ff()
{
  audio_state->p->ff();
}

void Audio::fb()
{
  audio_state->p->fb();
}

void Audio::vol_plus()
{
  audio_state->p->volup();
}

void Audio::vol_minus()
{
  audio_state->p->voldown();
}

void Audio::mute()
{
  audio_state->p->mute();
}

void Audio::toggle_playlist_helper()
{
  // FIXME: what is this function used for?
  change_mode = false;
  toggle_playlist();
}

bool Audio::check_mode(const Input& input, bool disable_mode_change)
{
  return (!disable_mode_change && (input.command == "mode" && input.mode == "audio_general") 
	  || change_mode);
}

bool Audio::check_playlist_mode(const Input& input, bool disable_mode_change)
{
  return (!disable_mode_change && input.command == "playlist_mode" && input.mode == "audio_general");
}

bool Audio::check_playlist_commands(const Input& input, bool disable_mode_change)
{
  return (!disable_mode_change && input.command == "playlist_commands");
}

bool Audio::check_update_time_on_keypress(const Input& input, bool disable_mode_change)
{
  return (input.command != "vol+" && input.command != "vol-" && 
	  input.command != "mute");
}

void Audio::update_time_on_keypress()
{
  conf->s_last_key(time(0));
}

void Audio::add_playback_keys()
{
  Playback p = Playback("Audio");
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::play), "play", gettext("Play")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::pause), "pause", gettext("Pause")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::stop), "stop", gettext("Stop")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::ff), "ff", gettext("Fast forward")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::fb), "fb", gettext("Fast backward")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::next_track), "next_track", gettext("Next track")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::prev_track), "prev_track", gettext("Previous track")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::vol_plus), "vol+", gettext("Increase volume")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::vol_minus), "vol-", gettext("Decrease volume")));
  p.add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::mute), "mute", gettext("Mute")));

  global->add_playback(p);
  global->set_playback("Audio");

  global->add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::toggle_playlist),
				    new InputCheck<Audio>(this, &Audio::check_mode)));
  
  global->add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::toggle_load_playlist),
				    new InputCheck<Audio>(this, &Audio::check_playlist_mode)));

  global->add_command(GlobalCommand(new TimeFunctorv<Audio>(this, &Audio::update_time_on_keypress),
				    new InputCheck<Audio>(this, &Audio::check_update_time_on_keypress)));

  global->add_second_menu_item(ExtraMenuItem(gettext("Toggle playlist"), 
					     input_master->find_shortcut("mode"),
					     new TimeFunctorv<Audio>(this, &Audio::toggle_playlist)));
}
