// used for #ifdef
#include "config.h"

#include "libfspp.hpp"

#include "game.hpp"

//Input
#include "input.hpp"

#include "common.hpp"
#include "gettext.hpp"
#include "graphics.hpp"

#include "common-feature.hpp"
#include "startmenu.hpp"
#include "extra_menu.hpp"
#include "updater.hpp"
#include "busy_indicator.hpp"
#include "shutdown.hpp"
#include "search.hpp"
#include "notify.hpp"

// ostringstream
#include <sstream>

#include <iostream>

#include <cassert>

using std::list;
using std::vector;
using std::string;

Game::GameOpts::GameOpts() 
{
  // header
  header = gettext("Game Options");

  // values

  // reload
  vector<string> none;
  reload_p = new Option(false, gettext("reload"), "reload", 0, none, none);
  val.push_back(reload_p);

  // sorting order
  val.push_back(go->dir_order);
}

Game::GameOpts::~GameOpts() 
{
  save();

  delete reload_p;
}

void Game::GameOpts::read_dirs() 
{ 
  Game *game = get_class<Game>("Game");
  game->read_dirs(); 
  game->reset();
}

Game::Game() 
 : db((conf->p_var_data_dir() + "games.db").c_str()), 
   exit_loop(false)
{
  if (conf->p_v_res() == 720)
    y_start = 85;
  else if (conf->p_v_res() == 800)
    y_start = 90;
  else if (conf->p_v_res() == 1024)
    y_start = 110;
  else if (conf->p_v_res() == 1280)
    y_start = 110;
  else
    y_start = 70; // default value

  if (conf->p_h_res() == 405)
    y_start = 70;

  image_width = 115;
  image_height = 120;

  if (conf->p_v_res() == 720)
    image_height = 105;

  if (conf->p_h_res() == 405) {
    image_width = 65;
    image_height = 67;
  }

  image_width_all = image_width+25;
  image_height_all = image_height+55;
  image_height_all_eks_text = image_height + 6;

  images_per_row = (conf->p_v_res()-2*25)/image_width_all;

  // lets make some more room for text 
  image_width_all += ((conf->p_v_res()-2*25)%image_width_all)/images_per_row;

  // + 20 is for the last row, since we don't need to add extra space
  rows = (conf->p_h_res()-2*y_start+20)/image_height_all;
  rows_search = (conf->p_h_res()-2*y_start+20-40)/image_height_all;

  image_height_all = (conf->p_h_res()-2*y_start+20)/rows;
  image_height_all_search = (conf->p_h_res()-2*y_start+20-40)/rows_search;

  game_conf = S_GameConfig::get_instance();

  check_db();

  reset();
}

void Game::startup_updater()
{
  EnterFunctor<Game> *e = new EnterFunctor<Game>(this, &Game::reenter);
  SearchGenList<Game> *s = new SearchGenList<Game>(this, &Game::gen_search_list);
  S_Search::get_instance()->register_module(SearchModule("Games", s, e));
}

void Game::create_db()
{
  vector<string> fields;
  fields.push_back("version INTEGER NOT NULL");
  db.createTable("Schema", fields);

  db.execute("INSERT INTO Schema VALUES('1')");

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("parent INTEGER");
  fields.push_back("filename TEXT");
  fields.push_back("name TEXT");
  fields.push_back("is_folder INTEGER"); // 1 = true
  db.createTable("Folders", fields);
}

void Game::check_db()
{
  if (db.hasTable("Schema")) {
    // check for correct version
    SQLQuery *q = db.query("Schema", "SELECT version FROM %t");
    if (q && q->numberOfTuples() > 0) {
      SQLRow &row = (*q)[0];
      if (row["version"] != "1")
	; // for upgrading db
    }
    delete q;
  } else {
    // create the whole db
    create_db();
  }
}

void Game::reset()
{
  for (int i = 0, size = folders.size(); i < size; ++i)
    folders.pop();

  list<string> dirs = game_conf->p_game_dirs();

  folders.push(std::make_pair(dirs, 0));
}

bool Game::file_sort::operator()(const GameEntry& lhs, const GameEntry& rhs)
{
  return sort_order(lhs.lowercase_name, rhs.lowercase_name, !lhs.path.empty(), !rhs.path.empty());
}

// commands
void Game::action_play()
{
  GameEntry e = vector_lookup(files, folders.top().second);
  int nfiles = e.filenames.size();
  if (nfiles > 0)
    playgame(e);
}

void Game::go_back()
{
  if (!(folders.size() == 1)) {
    folders.pop();
    load_current_dirs();

#ifdef use_inotify
    S_Notify::get_instance()->register_plugin("game", folders.top().first,
					      new TimeFunctorv<Game>(this, &Game::reparse_current_dir));
#endif
  }
}

void Game::enter_dir()
{
  if (vector_lookup(files, folders.top().second).type != "file")
    {
      string temp_cur_dir = vector_lookup(files ,folders.top().second).path;
      vector<GameEntry> templist = rdir(temp_cur_dir);

      if (templist.size() != 0) {
	list<string> templs; templs.push_back(temp_cur_dir + "/");
	folders.push(std::make_pair(templs, 0));
	files = templist;
#ifdef use_inotify
      S_Notify::get_instance()->register_plugin("game", folders.top().first,
						new TimeFunctorv<Game>(this, &Game::reparse_current_dir));
#endif
      } else {
	DialogWaitPrint pdialog(gettext("Folder is empty"), 1000);
      }
    }
}

// notify
void Game::reparse_current_dir()
{
  // process audio dir 
  load_current_dirs();

  // sanity check
  if (folders.top().second > files.size()-1)
    folders.top().second = 0;
}


void Game::load_current_dirs()
{
  files = parse_dir(folders.top().first);

  if (folders.top().first.size() > 1 && folders.size() == 1)
    std::sort(files.begin(), files.end(), file_sort());
}

void Game::options()
{
  BusyIndicator::get_instance().idle();
  
  if (opts.mainloop()) {
    // reload
    folders.top().second = 0;
  }
}

void Game::reenter(int db_id)
{
  reset();

  SQLQuery *q = db.query("Folders", ("SELECT * FROM %t WHERE id='" + conv::itos(db_id) + "'").c_str());

  if (q && q->numberOfTuples() > 0) {

    string filename = (*q)[0]["filename"];
    string cur_dir_path;

    if (filename[filename.size() - 1] == '/')
      filename = filename.substr(0, filename.size() - 1);

    string original_filename = filename;

    list<string> dirs = game_conf->p_game_dirs();

    int pos;
    for (list<string>::iterator i = dirs.begin(), iend = dirs.end(); i != iend; ++i)
      if ((pos = filename.rfind(*i)) != string::npos) {
	cur_dir_path = filename.substr(0, i->size());
	filename = filename.substr(i->size());
      }

    if (!filename.empty()) {
      while ((pos = filename.find("/")) != string::npos) {
	folders.top().second = find_position_in_folder_list(folders.top().first, cur_dir_path + filename.substr(0, pos + 1));

	list<string> templs; templs.push_back(cur_dir_path + filename.substr(0, pos + 1));
	folders.push(std::make_pair(templs, 0));
	cur_dir_path += filename.substr(0, pos+1);
	filename = filename.substr(pos+1);
      }
    }

    // we should now be in the correct folder, find the correct file
    folders.top().second = find_position_in_folder_list(folders.top().first, original_filename);

    load_current_dirs();

#ifdef use_inotify
    S_Notify::get_instance()->register_plugin("game", folders.top().first,
					      new TimeFunctorv<Game>(this, &Game::reparse_current_dir));
#endif

    mainloop();
  }

  delete q;
}

int Game::find_position_in_folder_list(const list<string>& files, 
				       const string& filename)
{
  int pos = 0;
  bool pos_found = false;
  for (list<string>::const_iterator i = files.begin(), iend = files.end(); i != iend; ++i){
    vector<GameEntry> templist = rdir(*i);
    pos = 0;
    for (vector<GameEntry>::iterator fi = templist.begin(), fiend = templist.end(); 
	 fi != fiend; ++fi, ++pos) {
      if (filename == fi->path ||
	  (fi->filenames.size() > 0 && filename == fi->filenames.front())) {
	pos_found = true;
	break;
      }
    }
    if (pos_found)
      break;
  }
  return pos;
}

vector<std::pair<string, int> > Game::gen_search_list(const string& search_word)
{
  std::vector<std::pair<string, int> > result;

  SQLQuery *q = db.query("Folders", ("SELECT DISTINCT id, filename, name, is_folder FROM %t WHERE fuzzycmp('" + search_word + "', name, 1)").c_str());

  if (q) {

    for (int i = 0; i < q->numberOfTuples(); ++i) {

      SQLRow &row = (*q)[i];

      string out;

      bool result_ok = false;
      list<string> current_folders = folders.top().first;
      for (list<string>::iterator i = current_folders.begin(), iend = current_folders.end();
	   i != iend; ++i) {
	if (row["filename"].find(*i) != string::npos) {
	  result_ok = true;
	  break;
	}
      }

      if (result_ok) {
	out = row["name"];
	result.push_back(std::make_pair(string_format::convert(out), conv::atoi(row["id"])));
      }
    }
  }
  delete q;

  return result;
}

void Game::search_func()
{
  BusyIndicator::get_instance().idle();

  ConstFunctor<Game, std::string, vector<std::pair<string, int> > > gen_list_func(this, &Game::gen_search_list);
  ConstFunctor<Game, vector<GameEntry>, void> print_func(this, &Game::graphical_print);

  int pos = input_master->device->graphical_search(files, &print_func, &gen_list_func,
						   search_mode, search_str, lowercase_search_str, 
						   offset, search_help_offset);

  if (pos != -1)
    folders.top().second = pos;
}

void Game::exit()
{
  exit_loop = true;
}

string Game::mainloop()
{
  BusyIndicator &busy_indicator = BusyIndicator::get_instance();
  Shutdown *sd = S_Shutdown::get_instance();
  
  Input input;

  bool update_needed = true;

  while (!exit_loop)
    {
      if (update_needed)
	graphical_print(files);

      busy_indicator.idle();

      input = input_master->device->get_input();

      busy_indicator.busy();

      if (sd->is_enabled()) {
      	sd->cancel();
      	continue;
      }

      update_needed = true;

      if (input.command == "prev")
	{
	  if (files.size() > images_per_row) {
	    for (int i = 0; i < images_per_row; ++i) {
	      if (folders.top().second == 0) {
		folders.top().second = files.size()-1;
		i += images_per_row-files.size()%images_per_row;
	      }
	      else 
		folders.top().second -= 1;
	    }
	  }
 	}
      else if (input.command == "next")
 	{
	  if (files.size() > images_per_row) {
	    for (int i = 0; i < images_per_row; ++i) {
	      if (folders.top().second == files.size()-1 && i == 0) {
		folders.top().second = 0;
		break;
	      } else if (folders.top().second == files.size()-1) {
		break;
	      } else {
		folders.top().second += 1;
	      }
	    }
	  }
 	}
      else if (input.command == "left")
	{
	  if (folders.top().second != 0)
	    folders.top().second -= 1;
	  else
	    folders.top().second = files.size() - 1;
	}
      else if (input.command == "right")
 	{
	  folders.top().second = (folders.top().second+1)%files.size();
 	}
      else if (input.command == "back" && folders.size() > 1)
	{
	  go_back();
	}
      else if (input.command == "back" && folders.size() == 1)
	{
	  exit();
	}
      else if (input.command == "play_game")
	{
	  action_play();
	}
      else if (input.command == "action")
	{
	  if (vector_lookup(files, folders.top().second).type != "file")
	    enter_dir();
	  else
	    action_play();
	}
      else if (input.command == "second_action")
	{
	  ExtraMenu em;

	  if (vector_lookup(files, folders.top().second).type != "file") {
	    em.add_item(ExtraMenuItem(gettext("Enter directory"), 
				      input_master->find_shortcut("action"), 
				      new TimeFunctorv<Game>(this, &Game::enter_dir)));
	  } else {
	    em.add_item(ExtraMenuItem(gettext("Play game"), 
				      input_master->find_shortcut("action"), 
				      new TimeFunctorv<Game>(this, &Game::action_play)));
	  }
	  
	  em.add_item(ExtraMenuItem(gettext("Search"), input_master->find_shortcut("search"), 
				    new TimeFunctorv<Game>(this, &Game::search_func)));
	  em.add_item(ExtraMenuItem(gettext("Options"), input_master->find_shortcut("options"), 
				    new TimeFunctorv<Game>(this, &Game::options)));

	  if (folders.size() > 1)
	    em.add_item(ExtraMenuItem(gettext("Go up one directory"), 
				      input_master->find_shortcut("back"), 
				      new TimeFunctorv<Game>(this, &Game::go_back)));
	  else {
	    em.add_item(ExtraMenuItem(gettext("Return to startmenu"), 
				      input_master->find_shortcut("back"), 
				      new TimeFunctorv<Game>(this, &Game::exit)));
	  }

	  // FIXME
	  //#ifdef c_music
	  //	  TimeFunctorv<Audio> t_player = TimeFunctorv<Audio>(Audio::get_instance(), &Audio::playlist_commands);
	  //	  em.add_item(ExtraMenuItem(gettext("Playlist commands"), input_master->find_shortcut("playlist_commands"), &t_player));
	  //#endif
	  em.mainloop();
	} 
      else if (input.command == "page_up")
 	{
	  int jump = conf->p_jump();

	  if (search_mode)
	    jump = images_per_row*rows_search;
	  else
	    jump = images_per_row*rows;

	  if (files.size() > jump) {
	    int diff = folders.top().second-jump;
	    if (folders.top().second == 0)
	      folders.top().second = (files.size()-1)+diff;
	    else if (diff < 0)
	      folders.top().second = 0;
	    else
	      folders.top().second = diff;
	  }
 	}
      else if (input.command == "page_down")
 	{
	  int jump = conf->p_jump();

	  if (search_mode)
	    jump = images_per_row*rows_search;
	  else
	    jump = images_per_row*rows;

	  if (files.size() > jump) {
	    if (folders.top().second > (files.size() - jump) && folders.top().second != files.size()-1)
	      folders.top().second = files.size()-1;
	    else
	      folders.top().second = (folders.top().second+jump)%files.size();
	  }
 	}
      else if (input.command == "options")
	{
	  options();
	}
      else if (input.command == "search")
	{
	  search_func();
        }

      update_needed = !global->check_commands(input);
    }

  exit_loop = false;

  return "";
}

string Game::get_name_from_file(const GameEntry& e)
{
  return e.name;
}

bool Game::search_compare(const GameEntry& s)
{
  return (s.lowercase_name.substr(0, search_str.size()) == lowercase_search_str);
}

// print a game element
void Game::print_game_element(const GameEntry& r, const GameEntry& position, int y)
{
  string name = r.name;

  if (r.type != "file")
    name += "/";

  string_format::format_to_size(name, "Vera/16", conf->p_v_res()-75-155, true);
  
  if (r == position) {
    render->current.add(new PObj(themes->game_marked_left, 70, y-1, 2, NOSCALING));
    render->current.add(new PObj(themes->game_marked, 70+25, y-1, conf->p_v_res()-(2*70)-50, 2));
    render->current.add(new PObj(themes->game_marked_right, 70+25+conf->p_v_res()-(2*70)-50, y-1, 2, NOSCALING));
  }
  
  std::ostringstream out; out << r.filenames.size();

  render->current.add(new TObj(name, "Vera/16", 75, y, 
			       themes->game_font1, themes->game_font2, 
			       themes->game_font3, 3));
  render->current.add(new TObj(r.filetype, "Vera/16", conf->p_v_res()-145, y, 
			       themes->game_font1, themes->game_font2, 
			       themes->game_font3, 3));
  render->current.add(new TObj(out.str(), "Vera/16", conf->p_v_res()-85, y, 
			       themes->game_font1, themes->game_font2, 
			       themes->game_font3, 3));
}

string Game::get_cover(const std::string& s)
{
  string path = "";
   string test = "";
  string::size_type pos;

  if ((pos = s.rfind(".")) != string::npos) {
    path = s.substr(0, pos);
    test = (path + ".png").c_str();
    if (file_exists(test))
      return test;
    test = (path + ".jpg").c_str();
    if (file_exists(test)) 
      return test;
  }
  return "";
}

void Game::graphical_print(const vector<GameEntry>& cur_files)
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->game_background, 0, 0, 0, SCALE_FULL));
  
  render->current.add(new PObj(themes->startmenu_game_dir, 25, 10, 2, NOSCALING));
  
  if (themes->show_header) {
    render->current.add(new TObj(gettext("Games"), "Vera/28", 100, 10, 
				 themes->game_header_font1, themes->game_header_font2, 
				 themes->game_header_font3, 2));
  }
  
  if (search_mode) {

    int y = 60;
    if (conf->p_v_res() == 720) {
      y -= 3;
    }

    render->current.add(new RObj(0, y, conf->p_v_res(), 45, 
				 themes->general_search_rectangle_color1, 
				 themes->general_search_rectangle_color2, 
				 themes->general_search_rectangle_color3, 
				 themes->general_search_rectangle_color4, 2));

    render->current.add(new PObj(themes->general_search, 47, y, 3, NOSCALING));

    render->current.add(new TObj(gettext("Search: ") + search_str, "Vera/22", 100, y+5, 
				 themes->game_font1, themes->game_font2, 
				 themes->game_font3, 3));
  }

  // print the range
  int pos, start, pos_in_list, cur_rows, cur_image_height_all;

  if (search_mode) {
    pos_in_list = offset;
    cur_rows = rows_search;
    cur_image_height_all = image_height_all_search;
  } else {
    pos_in_list = folders.top().second;
    cur_rows = rows;
    cur_image_height_all = image_height_all;
  }

  int nr_in_last_row = cur_files.size()%images_per_row;
  if (nr_in_last_row == 0)
    nr_in_last_row = images_per_row;

  if (cur_files.size() > cur_rows*images_per_row) {
    if (pos_in_list < std::ceil(double(cur_rows)/2.0)*images_per_row) {
      pos = pos_in_list;
      start = 0;
    } else if (pos_in_list >= cur_files.size() - (images_per_row+nr_in_last_row)) { // last two rows
      start = cur_files.size()-((cur_rows-1)*images_per_row+nr_in_last_row);
      pos = pos_in_list-start;
    } else {
      pos = pos_in_list%images_per_row+images_per_row;
      if (pos == images_per_row-1)
	pos = 2*images_per_row-1;
      start = pos_in_list-pos;
    }
  } else {
    pos = pos_in_list;
    start = 0;
  }

  int x_start, x = (conf->p_v_res()-images_per_row*image_width_all)/2, y = y_start;
  x_start = x;
  
  if (search_mode)
    y = y_start + 40;

  for (int j = 0; j < cur_rows*images_per_row && j < cur_files.size(); ++j) {

    GameEntry s = vector_lookup(cur_files, start);
    string name = s.name;

    // Ignoring rom not playable, only necesary for others roms
    if (s.name == "neogeo") 
      continue;
    
    if (s.type != "file") {

      PObj *p;
      
      if (pos == j)
	p = new PObj(themes->game_dir_marked, x, y, 3, NOSCALING);
      else
	p = new PObj(themes->game_dir, x, y, 3, NOSCALING);

      p->x = p->real_x = x + (image_width_all - p->w)/2;
      p->y = p->real_y = y + (image_height_all_eks_text - p->h)/2;
      
      render->current.add(p);

    } else {
    	string cover_name = "";
    	list<string>::const_iterator iter = (s.filenames).begin();
    	cover_name = get_cover(*iter);
    	 
      if (!file_exists(cover_name)) {
	PObj *p = new PObj(themes->game_mcover, x-5, y-5, 3, NOSCALING);
	p->x = p->real_x = x + (image_width_all - (p->w + 10))/2;
	p->y = p->real_y = y + (image_height_all_eks_text - (p->h + 10))/2;
	render->current.add(p);
      } else {

	bool border = false;
	
	if (pos == j)
	  border = true;

	if (!file_exists(render->image_get_fullpath(cover_name, image_width, 
						    image_height))) {
	  render->create_scaled_image_wrapper(cover_name, image_width, image_height);
	  // to insert the image into the transmap so that it won't fail when
	  // we draw the screen again
	  bool tmp;
	  render->image_dimensions(render->image_get_fullpath(cover_name, image_width, image_height), &tmp);
	}

	PFObj *p = new PFObj(cover_name, x, y, image_width, image_height, border, 3);
	
	p->x = p->real_x = x + (image_width_all - (p->w + 10))/2;
	p->y = p->real_y = y + (image_height_all_eks_text - (p->h + 10))/2;

	if (p->border) {
	  p->real_x -= 5;
	  p->real_y -= 5;
	} else {
	  p->real_x -= 1;
	  p->real_y -= 1;
	}

	render->current.add(p);
      }
    }

    string old_name = name;
    
    string_format::format_to_size(name, "Vera/13", image_width_all-10, false);

    string name_second_part = old_name.substr(name.size());

    if (name_second_part.size() > 0 && name_second_part[0] != ' ') {
      // splitting words
      string::size_type last_whole_word = name.rfind(' ');
      if (last_whole_word != string::npos) {
	name = name.substr(0, last_whole_word);
	name_second_part = old_name.substr(name.size());
      }
    }

    int x_size = string_format::calculate_string_width(name, "Vera/13");


    string_format::format_to_size(name_second_part, "Vera/13", image_width_all-10, false);

    int x_size2 = string_format::calculate_string_width(name_second_part, "Vera/13");

    if (pos == j) {
      render->current.add(new TObj(name, "Vera/13", x + (image_width_all-x_size)/2, y+image_height_all_eks_text, 
				   themes->game_marked_font1, themes->game_marked_font2, 
				   themes->game_marked_font3, 3));
      render->current.add(new TObj(name_second_part, "Vera/13", x + (image_width_all-x_size2)/2, y+image_height_all_eks_text+20, 
				   themes->game_marked_font1, themes->game_marked_font2, 
				   themes->game_marked_font3, 3));
    } else {
      render->current.add(new TObj(name, "Vera/13", x + (image_width_all-x_size)/2, y+image_height_all_eks_text, 
				   themes->game_font1, themes->game_font2, 
				   themes->game_font3, 3));
      render->current.add(new TObj(name_second_part, "Vera/13", x + (image_width_all-x_size2)/2, y+image_height_all_eks_text+20, 
				   themes->game_font1, themes->game_font2, 
				   themes->game_font3, 3));
    }
    
    x += image_width_all;

    if (start == cur_files.size()-1) {
      if ((j+1)%images_per_row != 0)
	j += (images_per_row-(j+1)%images_per_row);
      
      x = x_start;
      y += cur_image_height_all;

      start = 0;
      continue;
      
    } else
      ++start;
    
    if ((j+1)%images_per_row == 0) {
      y += image_height_all;
      x = x_start;
    }
  }

  std::ostringstream position;

  if (search_mode)
    if (cur_files.size() > 0)
      position << offset + 1 << "/" << cur_files.size();
    else
      position << "";
  else
    position << folders.top().second+1 << "/" << cur_files.size();
  
  int x_size = string_format::calculate_string_width(position.str(), "Vera/18");
    
  render->current.add(new TObj(position.str(), "Vera/18", conf->p_v_res()-(60+x_size), 20, 
			       themes->game_font1, themes->game_font2, 
			       themes->game_font3, 3));

  render->draw("game-graphical");

  render->image_mut.leaveMutex();
}

void Game::playgame(const GameEntry file)
{
  DialogWaitPrint pdialog(gettext("Starting Game..."), 1000);

  string cur_files = "";
  
  for (list<string>::const_iterator iter = (file.filenames).begin(); iter != (file.filenames).end(); ++iter)
       cur_files += " \"" + (*iter) + "\"";

  string cur_dir = file.filenames.front().substr(0,file.filenames.front().rfind("/")+1);

  // support playlist files
  if (file.filetype == "mame" || file_exists(cur_dir + "mame_roms")) {
    run::exclusive_external_program(game_conf->p_mame_path() + " " + game_conf->p_mame_opts() + ' ' + cur_files);
  } else if (file.filetype == "snes" || file_exists(cur_dir + "snes_roms")) {
    run::exclusive_external_program(game_conf->p_snes_path() + " " + game_conf->p_snes_opts() + ' ' + cur_files);
  } else if (file.filetype == "nes" || file_exists(cur_dir + "nes_roms")) {
    run::exclusive_external_program(game_conf->p_nes_path() + " " + game_conf->p_nes_opts() + ' ' + cur_files);
  } else if (file.filetype == "gens" || file_exists(cur_dir + "gens_roms")) {
    run::exclusive_external_program(game_conf->p_gens_path() + " " + game_conf->p_gens_opts() + ' ' + cur_files);
  } else if (file.filetype == "sh" || file_exists(cur_dir + "sh_scripts")) {
    run::exclusive_external_program(cur_files);
  }
}

#ifdef use_media
// play everything on a cd
void Game::showcd(int count)
{
    vector<GameEntry> backup_files;

    backup_files = files;

    std::stack<std::pair<list<string>, int > > backup_folders = folders;

    list<string> dirs;
    dirs.push_back(cd->get_mount_point());

    files = parse_dir(dirs);

    while (!folders.empty())
      folders.pop();

    folders.push(std::make_pair(dirs, 0));

    mainloop();

    files = backup_files;
    folders = backup_folders;

  //umount the disc
  run::external_program("umount " + cd->get_mount_point());
}

void Game::playpsx()
{
  string device = cd->get_device();

  //play disc
  if (game_conf->p_psxemu_enable()) {
    run::my_system((game_conf->p_psxemu_path() + ' ' + game_conf->p_psxemu_opts()).c_str());
  } else {
    DebugPrint perror(gettext("No psx emulator enabled in config file"),
		      Print::DEBUGGING, DebugPrint::CRITICAL, "GAME");
    return;
  }
}
#endif


// ====================================================================
//
// dir
//
// ====================================================================

// Add a file to the list
GameEntry Game::addfile(const string& name, const MyPair& filetype)
{
  GameEntry r;
  r.id = ++id;
  r.filenames.push_back(name);
  string tempname = name.substr(0, name.size()-((filetype.first).size()+1));
  string::size_type pos = tempname.rfind("/");
  if (pos != string::npos)
    r.name = tempname.substr(pos+1);
  else
    r.name = tempname;
  if (conf->p_convert())
    r.name = string_format::convert(r.name);
  r.lowercase_name = string_format::lowercase(r.name);
  r.type = "file";
  r.filetype = filetype.second;

  return r;
}

// Add a game dir as "one file" to the game list
GameEntry Game::add_dir(const string& f)
{
  GameEntry output;

  string filename = (f[f.size()-1] == '/') ? f.substr(0, f.size()-1) : f;

  string type = "";

  vector<string> known_filetypes;

  for (file_iterator<file_t, default_order> i (filename); i != i.end (); i.advance(false))
    {
      string name = i->getName();
      
      if (check_type(name, game_conf->p_filetypes_g()) != emptyMyPair) {
	known_filetypes.push_back(name);
	type = "rom";
      }
    }

  if (!known_filetypes.empty()) {
    output.id = ++id; 
    string::size_type pos = filename.rfind("/");
    output.name = filename.substr(pos+1);
    output.path = filename;

    output.type = "dir-single";
    if (known_filetypes.size() > 1) {
      string first = known_filetypes.front();
      vector<string>::iterator i = known_filetypes.begin();
      ++i;
      for (vector<string>::iterator end = known_filetypes.end();
	   i != end; ++i) 
	if (string_format::difference(first, *i) > 1) {
	  output.type = "dir-multi";
	  break;
	}
    }

    if (conf->p_convert())
      output.name = string_format::convert(output.name);

    const MyPair filetype = check_type(known_filetypes.front(), game_conf->p_filetypes_g());
    output.filetype = filetype.second; 
    for (vector<string>::iterator i = known_filetypes.begin(), end = known_filetypes.end();
	 i != end; ++i)
      output.filenames.push_back(*i);

  } else {

    output.id = ++id;
    string::size_type pos = filename.rfind("/");
    output.name = filename.substr(pos+1);
    output.type = "dir"; 
    output.path = filename;
    output.filetype = "dir";

  } 

  output.lowercase_name = string_format::lowercase(output.name);

  return output;
}

void Game::insert_file_into_db(const string& filename, const string& parent)
{
  string parent_id = "";
  
  string good_parent = string_format::unique_folder_name(parent);
  
  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_string(good_parent) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    assert(q->numberOfTuples() == 1);
    SQLRow &row = (*q)[0];
    parent_id = row["id"];
  }
  delete q;

  if (parent_id.empty()) {
    char *tmp = sqlite3_mprintf("INSERT INTO Folders VALUES(NULL, '0', '%q', '', '1')", good_parent.c_str());
    db.execute(tmp);
    sqlite3_free(tmp);

    q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_string(good_parent) + "'").c_str());
    if (q && q->numberOfTuples() > 0) {
      assert(q->numberOfTuples() == 1);
      SQLRow &row = (*q)[0];
      parent_id = row["id"];
    }
    delete q;
  }

  bool is_dir = isDirectory(filename);

  string good_filename;
  string cut_filename;
  
  if (is_dir) {
    good_filename = string_format::unique_folder_name(filename);
    cut_filename = good_filename.substr(0, good_filename.size()-1);
  } else {
    good_filename = filename;
    cut_filename = filename;
  }

  string::size_type i;
  if ((i = cut_filename.rfind('/')) == string::npos)
    i = 0;
  string name = cut_filename.substr((i != 0) ? i+1 : i);
  
  if (!is_dir && (i = name.rfind('.')) != string::npos)
    name = name.substr(0, i);

  char *tmp_insert = sqlite3_mprintf("INSERT INTO Folders VALUES(NULL, '%q', '%q', '%q', '%q')", 
				     parent_id.c_str(), good_filename.c_str(), name.c_str(),
				     conv::itos(is_dir).c_str());
  db.execute(tmp_insert);
  sqlite3_free(tmp_insert);
}

int Game::db_id(const string& file, bool is_dir)
{
  int i = -1;

  string good_file;

  if (is_dir)
    good_file = string_format::unique_folder_name(file);
  else
    good_file = file;

  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_string(good_file) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    SQLRow &row = (*q)[0];
    i = conv::atoi(row["id"]);
  }
  delete q;

  return i;
}

vector<string> Game::check_db_for_folders(const string& parent)
{
  vector<string> names;

  string good_parent = string_format::unique_folder_name(parent);

  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_string(good_parent) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    assert(q->numberOfTuples() == 1);
    SQLRow &row = (*q)[0];
    string parent_id = row["id"];
    SQLQuery *q2 = db.query("Folders", ("SELECT filename FROM %t WHERE parent='" + parent_id + "'").c_str());
    if (q2) {
      for (int i = 0; i < q2->numberOfTuples(); ++i) {
	SQLRow &row = (*q2)[i];
	names.push_back(row["filename"]);
      }
    }
    delete q2;
  }
  delete q;

  return names;
}

bool Game::rdir_internal(const string& filename, const string& argv, 
			 vector<GameEntry> &cur_files)
{
  if (global->check_stop_bit()) {
    cur_files.clear();
    return false;
  }

  if (isDirectory (filename)) { // dir
    GameEntry dir = add_dir(filename);
    if ((dir.filenames).size() != 0 || dir.filetype == "dir") {
      dir.db_id = db_id(dir.path, true);
      cur_files.push_back(dir);
    }
  } else { // file
    MyPair filetype = check_type(filename, game_conf->p_filetypes_g());
    if (filetype != emptyMyPair) {
      GameEntry file(addfile(filename, filetype));
      file.db_id = db_id(filename, false);
      cur_files.push_back(file);
    }
  }

  return true;
}

// read a dir and spit out a list of files and dirs. For games.
vector<GameEntry> Game::rdir(const string& argv)
{
  vector<GameEntry> cur_files;

  vector<string> db_files = check_db_for_folders(argv);
  
  if (db_files.size() > 0) {
    for (vector<string>::const_iterator i = db_files.begin(), iend = db_files.end();
	 i != iend; ++i)
      if (!rdir_internal(*i, argv, cur_files))
	break;
  } else {
    for (file_iterator<file_t, default_order> i (argv); i != i.end (); i.advance(false)) {
      string filename = filesystem::FExpand(i->getName());
      insert_file_into_db(filename, argv);
      if (!rdir_internal(filename, argv, cur_files))
	break;
    }
  }

  return cur_files;
}

vector<GameEntry> Game::parse_dir(const list<string>& dirs)
{
  vector<GameEntry> vec_files;

  for (list<string>::const_iterator iter = dirs.begin(); iter != dirs.end(); ++iter)
    {
      vector<GameEntry> tempfiles = rdir(*iter);

      for (vector<GameEntry>::const_iterator iter2 = tempfiles.begin(); iter2 != tempfiles.end(); ++iter2)
	vec_files.push_back(*iter2);
    }

  return vec_files;
}

void Game::read_dirs() 
{
  // process game dir 
  files = parse_dir(game_conf->p_game_dirs());
  
  if (game_conf->p_game_dirs().size() > 1)
    std::sort(files.begin(), files.end(), file_sort());
}

#ifdef use_media

void Game::determine_media()
{
 bool data_disc = cd->data_disc();
  cd->close();

  if (data_disc) {
    //mount the disc
    run::my_system_no_kill(("mount " + cd->get_mount_point()).c_str());
  
    int count;
    string mediatype = testdir(&count); 

    if (mediatype == "empty" || (mediatype == "psx" && !game_conf->p_psxemu_enable()))
      Print pdialog(gettext("The disc does not contain any recognizable files"),
		    Print::SCREEN);
    else if (count < 2)
      DialogWaitPrint pdialog(gettext("Starting Playback..."), 1000);

    if (mediatype != "roms")
      //umount the disc
      run::my_system_no_kill(("umount " + cd->get_mount_point()).c_str()); 
    
    if (mediatype == "roms")
      showcd(count);
    else if (mediatype == "psx" && game_conf->p_psxemu_enable())
      playpsx();
  } else {
    Print pdialog(gettext("The disc does not contain any recognizable files"),
		  Print::SCREEN);
  }
	
}

// help function for determine_media
string Game::testdir(int *count)
{
  *count = 0;

  for (file_iterator<file_t, default_order> i (cd->get_mount_point()); i != i.end (); ++i) 
    {
      string filename = filesystem::FExpand(i->getName());
      string cur_dir = string_format::lowercase(filename.substr(filename.rfind("/")+1));
	  
      if (isDirectory (filename)) {
		//++(*count);
      } else if (cur_dir == "system.cnf") { 
      	return "psx";
      } else {
	MyPair filetype = check_type(filename, game_conf->p_filetypes_g());
	if (filetype != emptyMyPair)
	  ++(*count);
      }
    }

  if (*count == 0)
    return "empty";
  else
    // none of the above found, it must be a divx
    return "roms";
}
#endif
