#include "mplayer.hpp"

#include "graphical_movie.hpp"
#include "simple_movie.hpp"

#include "global.hpp"

#include "renderer.hpp"

using std::string;

#define bsize 1024

Mplayer::Mplayer()
  : MoviePlayer(true, true, true, true), paused(false), running(false), in_fullscreen(false), 
    pos(0), stop_playback_ok(true), broken_pipe(false)
{}

bool Mplayer::wants_exclusivity()
{
  return !S_Render::get_instance()->device->supports_wid();
}

std::pair<bool, Pipe> Mplayer::play_movie_in_slave_mode(char *buffer, const string& sub_lang, 
							const string& audio_mode, const string& paths)
{
  return run::exclusive_external_program_pipe(movie_conf->p_mplayer_path() + 
					      " -noconsolecontrols -nojoystick -nolirc -nomouseinput "
					      " -nocache -slave -quiet -vo xv -osdlevel 0 -wid " + buffer + 
					      (sub_lang.empty() ? "" : " -slang " + sub_lang) + 
					      ' ' + audio_mode + ' ' + paths);
}

string Mplayer::get_audio_mode()
{
  string audio_mode = " -ao ";
  if (S_GlobalOptions::get_instance()->audio_device->pos == 0) {
    if (conf->p_first_dev_output() == "alsa")
      audio_mode += conf->p_first_dev_output() + ":device=" + conf->p_first_dev_path();
    else
      audio_mode += conf->p_first_dev_output() + ":" + conf->p_first_dev_path();
  } else {
    if (conf->p_second_dev_output() == "alsa")
      audio_mode += conf->p_second_dev_output() + ":device=" + conf->p_second_dev_path();
    else
      audio_mode += conf->p_second_dev_output() + ":" + conf->p_second_dev_path();
  }
  
  return audio_mode;
}

void Mplayer::play_movie(const string& paths, bool window)
{
  if (running) {
    pos = 0;
    stop_and_switch();
  }

  string audio_mode = get_audio_mode();
  string sub_lang = get_class<Movie>("Movie")->opts.sub_lang();

  std::pair<bool, Pipe> result;

  Render *render = S_Render::get_instance();

  if (render->device->supports_wid()) {

    char buffer[1024];

    if (window) {

      render->device->hide_fs_video();
      snprintf(buffer, 1024, "0x%x", (int) render->device->wid());
      result = play_movie_in_slave_mode(buffer, sub_lang, audio_mode, paths);
      render->device->show_window_video();
    
    } else {

      render->device->hide_window_video();
      snprintf(buffer, 1024, "0x%x", (int) render->device->fs_wid());
      result = play_movie_in_slave_mode(buffer, sub_lang, audio_mode, paths);
      render->device->show_fs_video();

    }

    checker = new PlaybackChecker(this);
    checker->start();

    setup_playback(result, window);

  } else {

    run::exclusive_external_program((movie_conf->p_mplayer_path() + " " +  movie_conf->p_mplayer_opts() +
				     (sub_lang.empty() ? "" : " -slang " + sub_lang) + 
				     ' ' + audio_mode + ((conf->p_input() != "lirc") ? "" :
							 (movie_conf->p_lircconf().empty() ? "" :
							  " -lircconf " + movie_conf->p_lircconf())) + 
				     ' ' + paths));
  }

  type = MOVIE;
  path = paths;
}

void Mplayer::setup_playback(const std::pair<bool, Pipe>& result, bool window)
{
  if (result.first) {
    Global *global = S_Global::get_instance();
    global->set_playback("Movie");
    p = result.second;
    broken_pipe = false;
    running = true;
    if (window)
      in_fullscreen = false;
    else
      in_fullscreen = true;

    Movie *mov = get_class<Movie>("Movie");

    if (window)
      mov->set_movie_playback(true);
    else
      mov->set_movie_playback(false);
  }
}

void Mplayer::play_disc(const string& play_cmd)
{
  string audio_mode = get_audio_mode();

  string sub_lang = get_class<Movie>("Movie")->opts.sub_lang();

  string command = play_cmd + movie_conf->p_mplayer_path() + ' ' + 
    movie_conf->p_mplayer_opts() + (sub_lang.empty() ? "" : " -slang " + sub_lang) + 
    ' ' + audio_mode +
    ((conf->p_input() != "lirc") ? "" : 
     (movie_conf->p_lircconf().empty() ? "" : " -lircconf " + movie_conf->p_lircconf() + ' ')) + ' ';

  run::exclusive_external_program(command);
}

void Mplayer::play_vcd(const string& device)
{
  string audio_mode = get_audio_mode();
  string vcdtitle = get_class<Movie>("Movie")->opts.vcdtitle();

  run::exclusive_external_program(movie_conf->p_mplayer_path() + ' ' + movie_conf->p_mplayer_dvd_opts() +
				  " vcd://" + vcdtitle + "/" + device + ' ' + audio_mode + 
				  ((conf->p_input() != "lirc") ? "" :
				   (movie_conf->p_lircconf().empty() ? "" :
				    " -lircconf " + movie_conf->p_lircconf())));
}

void Mplayer::play_dvd(const string& device)
{
  string dvdtitle, sub_lang, dvd_audio_lang, audio_mode = get_audio_mode();

  Movie *mov = get_class<Movie>("Movie");

  dvdtitle = mov->opts.dvdtitle();
  sub_lang = mov->opts.sub_lang();
  dvd_audio_lang = mov->opts.dvd_audio_lang();

  run::exclusive_external_program(movie_conf->p_mplayer_path() + ' ' + movie_conf->p_mplayer_dvd_opts() +
				  " dvd://" + dvdtitle + " -slang " + sub_lang +
				  " -dvd-device '" + device + "'" + " -alang " + 
				  dvd_audio_lang + ' ' + audio_mode +
				  ((conf->p_input() != "lirc") ? "" :
				   (movie_conf->p_lircconf().empty() ? "" :
				    " -lircconf " + movie_conf->p_lircconf())));
}

void Mplayer::play()
{
  if (paused) {
    send_command("pause");
    paused = false;
  }
}

void Mplayer::pause()
{
  send_command("pause");
  paused = !paused;
}

void Mplayer::stop()
{
  stop_and_switch();
  run::release_exclusive_plugin_access();
  pos = 0;
}

void Mplayer::stop_and_switch()
{
  send_command("quit");

  playback_ended();
}

void Mplayer::ff()
{
  send_command("seek 10 0");
}

void Mplayer::fb()
{
  send_command("seek -10 0");
}

bool Mplayer::fullscreen()
{
  return in_fullscreen;
}

bool Mplayer::playing()
{
  return running;
}

void Mplayer::switch_screen_size()
{
  send_command("get_time_pos");
  stop_playback_ok = false;
  while (!stop_playback_ok)
    usleep(100*1000);
  stop_and_switch();

  if (type == MOVIE)
    play_movie(path, in_fullscreen); // in_fullscreen works because play_movie takes window
}

void Mplayer::playback_ended()
{
  get_class<Movie>("Movie")->set_movie_playback(false);

  Global *global = S_Global::get_instance();
  global->set_playback("Audio");
  broken_pipe=true;
  running = false;
  Render *render = S_Render::get_instance();
  render->device->hide_video();
  run::close_pipe(p);
}

void Mplayer::send_command(const std::string& command)
{
  const char *buff = command.c_str();

  // check for writeable pipe i.e. prevent broken pipe signal
  if(!broken_pipe) {
    struct pollfd pfd;
    pfd.fd=p.inpipe1; pfd.events=POLLOUT; pfd.revents=0;
    int r=poll(&pfd,1,50);
    if(r>0) {
      if(pfd.revents & ~POLLOUT) {
	printf("mplayer: %s%s%s%sin MPlayerControl\n",pfd.revents&POLLOUT?"POLLOUT ":"",pfd.revents&POLLERR?"POLLERR ":"",pfd.revents&POLLHUP?"POLLHUP ":"",pfd.revents&POLLNVAL?"POLLNVAL ":"");
	playback_ended();
      }
      else if(pfd.revents & POLLOUT) {
	r=write(p.inpipe1,buff,strlen(buff));
	if(r<0) {
	  printf("mplayer: pipe write(1) failed\n");
	  playback_ended();
	}
	else {
	  r=write(p.inpipe1,"\n",1);
	  if(r<0) {
	    printf("mplayer: pipe write(2) failed\n");
	    playback_ended();
	  }
	}
      }
    }
    else if(r==0)
      printf("mplayer: poll timed out in MPlayerControl (hugh?)\n");
    else 
      printf("mplayer: poll failed in MPlayerControl\n");
  } else {
    printf("mplayer: cmd pipe is broken\n");
    playback_ended();
  }

  //printf("mplayer: slave cmd: %s\n", buff);
}

PlaybackChecker::PlaybackChecker(Mplayer* mplayer)
  : player(mplayer), stop(false)
{}

void PlaybackChecker::run()
{
  struct pollfd pfd;
  pfd.fd=player->p.outpipe0; pfd.events=POLLIN;

  char buff[bsize+2]; // additional space for fake newline
  int c=0;

  while (player->running) {
    poll(&pfd,1,300);
    int r=read(player->p.outpipe0,buff+c,bsize-c);
    if(r>0) c+=r;
    if(c>0) {
      buff[c]=0; // make sure buffer is NULL terminated
      char *p;
      do {
        p=strpbrk(buff,"\n\r");
        if(!p && c==bsize) { // Full buffer, but no newline found.
          p=&buff[c];        // Have to fake one.
          buff[c]='\n'; c++; buff[c]=0;
          }
        if(p) {
          *p++=0;
          float itime;
          if(strncmp(buff,"Starting playback",17)==0) {
	    if (player->pos > 0.0) {
	      char tmp[20];
	      sprintf(tmp, "seek %.1f 2", player->pos);
	      player->send_command(tmp);
	    }
	  }
          else if(sscanf(buff,"ANS_TIME_POSITION=%f",&itime)==1) {
	    player->stop_playback_ok = true;
            if(itime>0) {
	      player->pos=itime;
	    }
	  }
          c-=(p-buff);
          memmove(buff,p,c+1);
	}
      } while(c>0 && p);
    }
  }
}
