#include "pywindow.hpp"
#include "gui/guiwindowmanager.hpp"

using namespace pymms::gui;

#ifdef __cplusplus
extern "C" {
#endif

namespace pymms {

namespace python {

PyObject* Window_New(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  Window *self;

  self = (Window*)type->tp_alloc(type, 0);
  if (!self) 
    return NULL;

  // Python window start at 1001
  int id = 1001;
  GUIWindowManager *windowmanager = S_WindowManager::get_instance();
  while(windowmanager->get_window(id))
    id++; 

  self->guiwindow = new GUIPYWindow();
  self->guiwindow->set_id(id);

  windowmanager->register_window(self->guiwindow);

  return (PyObject*)self;
}

void Window_Dealloc(Window* self)
{
  for (std::vector<Control*>::iterator i = self->controls.begin(),
         end = self->controls.end(); i != end; i++) {

    self->guiwindow->remove_control((*i)->guicontrol);
    Py_DECREF(*i);
  }

  S_WindowManager::get_instance()->unregister_window(self->guiwindow);

  delete self->guiwindow;
  self->ob_type->tp_free((PyObject*)self);
}

PyObject* Window_close(Window *self, PyObject *args)
{
  self->is_open = false;

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject* Window_on_action(Window *self, PyObject *args)
{
  char *action;

  if (!PyArg_ParseTuple(args, "s", &action))
    return NULL;

  std::string temp(action);
  if(temp == "back")
    Window_close(self, args);

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject* Window_show(Window *self, PyObject *args)
{
  S_WindowManager::get_instance()->set_active_window(self->guiwindow);

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject* Window_do_modal(Window *self, PyObject *args)
{
  if(!self->is_open)
  {
    Window_show(self, NULL);
    self->is_open = true;
    while(self->is_open)
    {
      Py_BEGIN_ALLOW_THREADS
      self->guiwindow->wait_for_action_event();
      Py_END_ALLOW_THREADS
    
      self->guiwindow->make_pending_call((PyObject *)self);
    }
  }

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject* Window_add_control(Window *self, PyObject *args)
{
  Control* control;

  if (!PyArg_ParseTuple(args, "O", &control)) 
    return NULL;

  if(!Control_Check(control))
  {
    PyErr_SetString(PyExc_TypeError, "Object should be of type Control");
    return NULL;
  }

  if(!self->guiwindow->add_control(control->guicontrol))
  {
    PyErr_SetString(PyExc_ReferenceError, "Control is already used");
    return NULL;
  }

  Py_INCREF(control);
  
  self->controls.push_back(control);

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject* Window_set_focus(Window *self, PyObject *args)
{
  Control *control;

  if (!PyArg_ParseTuple(args, "O", &control)) 
    return NULL;

  if(!Control_Check(control))
  {
    PyErr_SetString(PyExc_TypeError, "Object should be of type Control");
    return NULL;
  }

  if(!self->guiwindow->set_focus(control->guicontrol))
  {
    PyErr_SetString(PyExc_RuntimeError, "Control does not exist in window");
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject* Window_get_focus(Window *self, PyObject *args)
{
  GUIControl *guicontrol;
  Control *control = 0;

  if(!(guicontrol = self->guiwindow->get_focus()))
  {
    PyErr_SetString(PyExc_RuntimeError, "No control in this window has focus");
    return NULL;
  }

  for (std::vector<Control*>::iterator i = self->controls.begin(),
      end = self->controls.end(); i != end; i++) {
    if((*i)->guicontrol == guicontrol)
      control = *i;
  }

  if(!control)
  {
    PyErr_SetString(PyExc_RuntimeError, "This error should not happen :-)");
    return NULL;
  }

  Py_INCREF(control);

  return (PyObject*)control;
}

PyObject* Window_get_width(Window *self, PyObject *args)
{ 
  return PyLong_FromLong(self->guiwindow->get_width());
}

PyObject* Window_get_height(Window *self, PyObject *args)
{
  return PyLong_FromLong(self->guiwindow->get_height());
}

PyObject* Window_remove_control(Window *self, PyObject *args)
{ 
  Control* control;
  if (!PyArg_ParseTuple(args, "O", &control)) 
    return NULL;
  // type checking, object should be of type Control
  if(!Control_Check(control))
  {
    PyErr_SetString(PyExc_TypeError, "Object should be of type Control");
    return NULL;
  }

  if(!self->guiwindow->remove_control(control->guicontrol))
  {
    PyErr_SetString(PyExc_RuntimeError, "Control does not exist in window");
    return NULL;
  }

  for (std::vector<Control*>::iterator i = self->controls.begin(),
      end = self->controls.end(); i != end; i++) {
    if((*i) == control)
      self->controls.erase(i);
  }

  Py_DECREF(control);

  Py_INCREF(Py_None);
  return Py_None;
}

PyMethodDef Window_methods[] = {
  {"on_action", (PyCFunction)Window_on_action, METH_VARARGS, 0},
  {"do_modal", (PyCFunction)Window_do_modal, METH_VARARGS, 0},
  {"close", (PyCFunction)Window_close, METH_VARARGS, 0},
  {"add_control", (PyCFunction)Window_add_control, METH_VARARGS, 0},
  {"set_focus", (PyCFunction)Window_set_focus, METH_VARARGS, 0},
  {"get_focus", (PyCFunction)Window_get_focus, METH_VARARGS, 0},
  {"get_height", (PyCFunction)Window_get_height, METH_VARARGS, 0},
  {"get_width", (PyCFunction)Window_get_width, METH_VARARGS, 0},
  {"remove_control", (PyCFunction)Window_remove_control, METH_VARARGS, 0},
  {"show", (PyCFunction)Window_show, METH_VARARGS, 0},
  {NULL, NULL, 0, NULL}
};

PyTypeObject Window_Type = {
	PyObject_HEAD_INIT(NULL)
	0,                         /*ob_size*/
	"mmsv2gui.Window",         /*tp_name*/
	sizeof(Window),            /*tp_basicsize*/
	0,                         /*tp_itemsize*/
	(destructor)Window_Dealloc,/*tp_dealloc*/
	0,                         /*tp_print*/
	0,                         /*tp_getattr*/
	0,                         /*tp_setattr*/
	0,                         /*tp_compare*/
	0,                         /*tp_repr*/
	0,                         /*tp_as_number*/
	0,                         /*tp_as_sequence*/
	0,                         /*tp_as_mapping*/
	0,                         /*tp_hash */
	0,                         /*tp_call*/
	0,                         /*tp_str*/
	0,                         /*tp_getattro*/
	0,                         /*tp_setattro*/
	0,                         /*tp_as_buffer*/
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
	0,      /* tp_doc */
	0,		                     /* tp_traverse */
	0,		                     /* tp_clear */
	0,		                     /* tp_richcompare */
	0,		                     /* tp_weaklistoffset */
	0,		                     /* tp_iter */
	0,		                     /* tp_iternext */
	Window_methods,            /* tp_methods */
	0,                         /* tp_members */
	0,                         /* tp_getset */
	0,                         /* tp_base */
	0,                         /* tp_dict */
	0,                         /* tp_descr_get */
	0,                         /* tp_descr_set */
	0,                         /* tp_dictoffset */
	0,                         /* tp_init */
	0,                         /* tp_alloc */
	Window_New,                /* tp_new */
};

}

}

#ifdef __cplusplus
}
#endif


