/* 
 *  Copyright (C) 2006, 2007 by:
 *  Da He <da.he@uni-dortmund.de>
 *  Julian Flake <julian.flake@uni-dortmund.de>
 *  Anders Rune Jensen <anders@gnulinux.dk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "cRenderer.hpp"

#include "common.hpp"

#include "plugins.hpp"
#include "audio_config.hpp"
#include "upnp_config.hpp"

#include <fstream>

using std::string;
using std::cout;
using std::cerr;
using std::endl;

void* pt2obj; 

cRenderer::cRenderer()
  : audioplayer(0)
{
  UpnpConfig *upnp_conf = S_UpnpConfig::get_instance();
  upnp_conf->parse_configuration_file(conf->p_homedir());
    
  pt2obj = (void*)this;
	
  friendlyName = "MMS";

  ip_addr = upnp_conf->p_ip();

  this->udn = strdup(this->get_uuid());

  AudioConfig *audio_conf = S_AudioConfig::get_instance();
  Plugins *plugins = S_Plugins::get_instance();

  for (std::vector<AudioPlayerPlugin*>::const_iterator iter=plugins->audio_players.begin(),
         iend=plugins->audio_players.end(); iter!=iend; ++iter)
    if ( (*iter)->plugin_name() == audio_conf->p_audio_player()) {
      audioplayer = (*iter)->player;
      break;
    }

  if (audioplayer == 0) {
    cerr << gettext("UPNP: music player not found") << std::endl;
    ::exit(1);
  }

  positionToSeek = 0;
	
  services[ConnectionManager] = new cConnectionManager(this);
  services[AVTransport] = new cAVTransport(this);
  services[RenderingControl] = new cRenderingControl(this);

  //mutex init
  ithread_mutex_init(&rendererMutex,NULL);

  web_root = conf->p_var_data_dir() + "/upnp/";

  if (!file_exists(web_root + "AVTransport.xml")) {
    cerr << gettext("could not find description file.") << endl;
    cerr << gettext("make sure you have run make install.") << endl;
    ::exit(1);
  }
	
  if(ip_addr.empty())
    {
      cout << "random ip" << endl;
    }
  else
    {
      cout << ip_addr << endl;
    }

  int err;
  err = UpnpInit(ip_addr.c_str(), 0);
  if (err != UPNP_E_SUCCESS) {
    cerr << gettext("could not init SDK: ") << UpnpGetErrorMessage(err) << endl;
    UpnpFinish();
    ::exit(1);
  }
  if (UpnpSetWebServerRootDir(web_root.c_str()) != UPNP_E_SUCCESS) {
    cerr << gettext("could not set web server's root dir.") << endl;
    UpnpFinish();
    ::exit(1);
  }
	
  IXML_Document *desc = getDescriptionDoc();
  string str = ixmlPrintDocument(desc);
  int lng = str.length();

  //	cout << ixmlPrintDocument(desc) << endl;
  err = UpnpRegisterRootDevice2(UPNPREG_BUF_DESC, str.c_str(), lng,
				1, callbackEventHandler, NULL, &renderer_hdl);
  if (err != UPNP_E_SUCCESS) {
    cerr << gettext("could not register root device: ") << UpnpGetErrorMessage(err) << endl;
    UpnpFinish();
    ::exit(1);
  }		
  cout << "SDK intitalized and device registered." << endl;

  // Advertisement
  if (UpnpSendAdvertisement(renderer_hdl, 1000) != UPNP_E_SUCCESS) {
    cerr << gettext("could not advertise device") << endl;
    UpnpFinish();
    ::exit(1);
  }		
  cout << "device advertised." << endl;
}

cRenderer::~cRenderer() 
{
  int err;
  err=UpnpUnRegisterRootDevice(renderer_hdl);
  if (err != UPNP_E_SUCCESS) {
    cerr << gettext("could not unregister root device:") << err << endl;
    UpnpFinish();
    ::exit(1);
  }		
  if (UpnpFinish() != UPNP_E_SUCCESS) {
    cerr << gettext("could not finish") << endl;
    ::exit(1);
  }

  pt2obj = 0;

  delete services[ConnectionManager];
  delete services[AVTransport];
  delete services[RenderingControl];

  ithread_mutex_destroy(&rendererMutex);
}

/*
 * Handles Action Requests
 */
void cRenderer::handleActionRequest(Upnp_Action_Request *Event)
{
  ithread_mutex_lock(&rendererMutex);
  cout << ixmlPrintDocument(Event->ActionRequest) << endl;	
  if ( (strcmp(Event->ServiceID,services[RenderingControl]->service_id)==0)&&
      (strcmp(Event->DevUDN,udn)==0)){
    for(int i=0;i<services[RenderingControl]->action_count;i++){
      if(!strcmp(Event->ActionName, services[RenderingControl]->action_name[i])){
	((cRenderingControl*)services[RenderingControl])->callAction(i,Event->ActionRequest,&Event->ActionResult,Event);
	break;
      }						   
    }
  } else if( (strcmp(Event->DevUDN,udn)==0)&&
	    (strcmp(Event->ServiceID,services[ConnectionManager]->service_id)==0)){
    for(int i=0;i<services[ConnectionManager]->action_count;i++){
      if(!strcmp(Event->ActionName, services[ConnectionManager]->action_name[i])){
	((cConnectionManager*)services[ConnectionManager])->callAction(i,Event->ActionRequest,&Event->ActionResult,Event);
	break;
      }						   
    }
  } else if( (strcmp(Event->DevUDN,udn)==0)&&
	    (strcmp(Event->ServiceID,services[AVTransport]->service_id)==0)){
    for(int i=0;i<services[AVTransport]->action_count;i++){
      if(!strcmp(Event->ActionName, services[AVTransport]->action_name[i])){
	((cAVTransport*)services[AVTransport])->callAction(i,Event->ActionRequest,
							   &Event->ActionResult,Event);
	break;
      }						   
    }
  } else{
    cout << "service false" << endl;
    return;
  }
	
  ithread_mutex_unlock(&rendererMutex);
}

/*
 * Handles Subscription Requests
 */
void cRenderer::handleSubscriptionRequest(Upnp_Subscription_Request* Event)
{
  ithread_mutex_lock(&rendererMutex);
  int err = 254;
  if (!strcmp("urn:upnp-org:serviceId:RenderingControl1",Event->ServiceId)) {
    cout << "subscription of service RenderingControl1 requested" << endl;
    char* evRC_name[1];
    char* evRC_value[1];
    evRC_name[0]=services[RenderingControl]->name[cRenderingControl::RC_LastChange] ;
    evRC_value[0]=services[RenderingControl]->value[cRenderingControl::RC_LastChange] ;
		
    cout << evRC_name[0] << endl;
    err = UpnpAcceptSubscription(renderer_hdl,
				 Event->UDN,
				 Event->ServiceId,
				 (const char **) evRC_name,
				 (const char **) evRC_value,
				 1,
				 Event->Sid);
    cout << "nach accept" << endl;
    (services[RenderingControl])->setVar(RenderingControl,1,cRenderingControl::RC_Volume,services[RenderingControl]->value[cRenderingControl::RC_Volume]);

  } else if (!strcmp("urn:upnp-org:serviceId:ConnectionManager1",Event->ServiceId)) {
    cout << "subscription of service ConnectionManager1 requested" << endl;
    char* evCM_name[3];
    char* evCM_value[3];
		
    evCM_name[0]=services[ConnectionManager]->name[cConnectionManager::CM_SourceProtocolInfo];
    evCM_name[1]=services[ConnectionManager]->name[cConnectionManager::CM_SinkProtocolInfo];
    evCM_name[2]=services[ConnectionManager]->name[cConnectionManager::CM_CurrentConnectionIDs];
		
    evCM_value[0]=services[ConnectionManager]->value[cConnectionManager::CM_SourceProtocolInfo];
    evCM_value[1]=services[ConnectionManager]->value[cConnectionManager::CM_SinkProtocolInfo];
    evCM_value[2]=services[ConnectionManager]->value[cConnectionManager::CM_CurrentConnectionIDs];
		
    err = UpnpAcceptSubscription(renderer_hdl,
				 Event->UDN,
				 Event->ServiceId,
				 (const char **) evCM_name,
				 (const char **) evCM_value,
				 3, // muss noch korrigiert werden
				 Event->Sid);
  } else if (!strcmp("urn:upnp-org:serviceId:AVTransport1",Event->ServiceId)) {
    cout << "subscription of service AVTransport1 requested" << endl;
    char* evAV_name[1];
    char* evAV_value[1];
    evAV_name[0]=services[AVTransport]->name[cAVTransport::AV_LastChange];
    evAV_value[0]=services[AVTransport]->value[cAVTransport::AV_LastChange];
    err = UpnpAcceptSubscription(renderer_hdl,
				 Event->UDN,
				 Event->ServiceId,
				 (const char **) evAV_name,
				 (const char **) evAV_value,
				 1,
				 Event->Sid);
  } else {		
    cout << "subscription of uncatched service " <<
      Event->ServiceId <<
      " requested" <<
      endl;
  }
  cout << "accepted? " << Event->ServiceId << " " ;	
  cout << " - err: " << err << endl;

	
  ithread_mutex_unlock(&rendererMutex);
}

/*
 * Handles Get Var Requests
 */
void cRenderer::handleGetVarRequest(Upnp_State_Var_Request* Event)
{
  ithread_mutex_lock(&rendererMutex);
	
  cout << "get var requested:" << 
    Event->DevUDN <<
    Event->ServiceID <<
    Event->StateVarName <<
    //			Event->CtrlPtIPAddr <<
    Event->CurrentVal <<
    endl;
	
  ithread_mutex_unlock(&rendererMutex);
}

/*
 * Handles sdk's callbacks
 */
int cRenderer::callbackEventHandler(Upnp_EventType EventType, void *Event, void *Cookie)
{
  cRenderer *cp = (cRenderer*) pt2obj;
  cout << "cp->UDN " << cp->udn << endl;

  switch (EventType) {
  case UPNP_CONTROL_ACTION_REQUEST: { /* Received by a device when a control point issues a control request. */
    cp->handleActionRequest((struct Upnp_Action_Request *) Event);
    break; 
  }
  case UPNP_CONTROL_GET_VAR_REQUEST: { /* Received by a device when a query for a single service variable arrives. */
    // deprecated
    cp->handleGetVarRequest((struct Upnp_State_Var_Request *) Event);
    break; 
  }
  case UPNP_EVENT_SUBSCRIPTION_REQUEST: { /* Received by a device when a query for a single service variable arrives. */
    cp->handleSubscriptionRequest((struct Upnp_Subscription_Request *) Event);
    break; 
  }
  default: {
    cout << "uncatched Event: " << EventType << endl;
    break;
  }
  }
  return UPNP_E_SUCCESS;
}

void cRenderer::wakeup(int in)
{
  cRenderer* cp = (cRenderer*) pt2obj;
  if (strcmp(cp->services[AVTransport]->value[cAVTransport::AV_TransportState],"STOPPED")!=0) {
    cp->services[AVTransport]->setVar(AVTransport,2,cAVTransport::AV_TransportState,"STOPPED");
    // notification
    IXML_Document* LCdoc = NULL;
    IXML_Node* node = NULL;
    IXML_Element* Ele = NULL;
    LCdoc = ixmlParseBuffer("<Event xmlns=\"urn:schemas-upnp-org:metadata-1-0/AVT/\"></Event>");
    node = ixmlNode_getFirstChild( (IXML_Node*) LCdoc );
    Ele = ixmlDocument_createElement( LCdoc, "InstanceID" );
    ixmlElement_setAttribute(Ele, "val","0");
    ixmlNode_appendChild(  node,(IXML_Node*) Ele);
    IXML_Element* Ele2 = NULL;
    Ele2 = ixmlDocument_createElement(LCdoc, 
				      cp->services[AVTransport]->name[cAVTransport::AV_TransportState]);
    ixmlElement_setAttribute(Ele2, "val", cp->services[AVTransport]->value[cAVTransport::AV_TransportState]);
    ixmlNode_appendChild((IXML_Node*)Ele,(IXML_Node*) Ele2);
    IXML_Document* myDoc2 = UpnpCreatePropertySet(1,
						  "LastChange",
						  ixmlPrintDocument(LCdoc)
						  ); 
    UpnpNotifyExt(cp->renderer_hdl,cp->udn,cp->services[AVTransport]->service_id,myDoc2);
    cout << "SIGCHLD" << endl;
  }
}

void cRenderer::seconds2tstamp(char* out, const char* in)
{
	int ihours = 0;
	int iminutes = 0;
	int iseconds = atoi(in);
	
	iminutes = iseconds/60;
	iseconds = iseconds - iminutes*60;
	ihours = iminutes/60;
	iminutes = iminutes - ihours*60;
	
	char chours[3];
	char cminutes[3];
	char cseconds[3];
	
	if (ihours<10) sprintf(chours,"0%i",ihours);
	else sprintf(chours,"%i",ihours);
	
	if (iminutes<10) sprintf(cminutes,"0%i",iminutes);
	else sprintf(cminutes,"%i",iminutes);
	
	if (iseconds<10) sprintf(cseconds,"0%i",iseconds);
	else sprintf(cseconds,"%i",iseconds);
	
	sprintf(out,"%s:%s:%s",chours,cminutes,cseconds);
}

int cRenderer::time2seconds(char* in)
{
	string length = in;

	string cHours = length.substr(0,length.find_first_of(':'));

	length = length.substr(length.find_first_of(':')+1);
	string cMinutes = length.substr(0,length.find_first_of(':'));

	length = length.substr(length.find_first_of(':')+1);
	string cSeconds = length.substr(0,length.find_first_of(':'));
	
	int iHours = atoi(cHours.c_str());
	int iMinutes = atoi(cMinutes.c_str());
	int iSeconds = atoi(cSeconds.c_str());
	
	iSeconds = iSeconds + (60*iMinutes) + (3600*iHours);
	return iSeconds;
}

IXML_Document* cRenderer::getDescriptionDoc()
{
  IXML_Document* descDoc;
  IXML_Node* node;
  IXML_Node* node2;
  IXML_Node* node3;
  IXML_Element* Ele;
  IXML_Element* Ele2;
  IXML_Element* Ele3;
  IXML_Element* Ele4;
  descDoc = ixmlParseBuffer("<root></root>");
  node = ixmlNode_getFirstChild( (IXML_Node*) descDoc );
	
  Ele = ixmlDocument_createElement( descDoc, "specVersion" );
  Ele2 = ixmlDocument_createElement( descDoc, "major" );
  node2 = ixmlDocument_createTextNode( descDoc, "1" );
  ixmlNode_appendChild((IXML_Node*) Ele2, (IXML_Node*) node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);
  Ele2 = ixmlDocument_createElement( descDoc, "minor" );
  node2 = ixmlDocument_createTextNode( descDoc, "0" );
  ixmlNode_appendChild((IXML_Node*) Ele2, (IXML_Node*) node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);
  ixmlNode_appendChild(node, (IXML_Node*) Ele);

  // Beginn <device>
  Ele = ixmlDocument_createElement( descDoc, "device" );

  Ele2 = ixmlDocument_createElement( descDoc, "deviceType" );
  node2 = ixmlDocument_createTextNode(descDoc, "urn:schemas-upnp-org:device:MediaRenderer:1");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);
	
  Ele2 = ixmlDocument_createElement( descDoc, "friendlyName" );
  node2 = ixmlDocument_createTextNode(descDoc, friendlyName);
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);
	
  Ele2 = ixmlDocument_createElement( descDoc, "manufacturer" );
  node2 = ixmlDocument_createTextNode(descDoc, "PG483");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "manufacturerURL" );
  node2 = ixmlDocument_createTextNode(descDoc, "http://www.ubiquitous-media.de");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "modelDescription" );
  node2 = ixmlDocument_createTextNode(descDoc, "Linux-Renderer(C++,PG483) 1.0");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "modelName" );
  node2 = ixmlDocument_createTextNode(descDoc, "AV Renderer");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "modelNumber" );
  node2 = ixmlDocument_createTextNode(descDoc, "1.0");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "modelURL" );
  node2 = ixmlDocument_createTextNode(descDoc, "http://www.ubiquitous-media.de");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "serialNumber" );
  node2 = ixmlDocument_createTextNode(descDoc, "000000000001");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "UDN" );
  node2 = ixmlDocument_createTextNode(descDoc, udn);
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "UPC" );
  node2 = ixmlDocument_createTextNode(descDoc, "000000000001");
  ixmlNode_appendChild((IXML_Node*) Ele2, node2);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele2 = ixmlDocument_createElement( descDoc, "iconList" );
  Ele3 = ixmlDocument_createElement(descDoc, "icon");
  Ele4 = ixmlDocument_createElement(descDoc, "mimeType");
  node3 = ixmlDocument_createTextNode(descDoc, "image/gif");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "width");
  node3 = ixmlDocument_createTextNode(descDoc, "32");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "height");
  node3 = ixmlDocument_createTextNode(descDoc, "32");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "depth");
  node3 = ixmlDocument_createTextNode(descDoc, "8");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "url");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnpmplayer.gif");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  ixmlNode_appendChild((IXML_Node*) Ele2, (IXML_Node*) Ele3);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

	
  Ele2 = ixmlDocument_createElement( descDoc, "serviceList" );
	
  Ele3 = ixmlDocument_createElement(descDoc, "service");
  Ele4 = ixmlDocument_createElement(descDoc, "serviceType");
  node3 = ixmlDocument_createTextNode(descDoc, "urn:schemas-upnp-org:service:RenderingControl:1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "serviceId");
  node3 = ixmlDocument_createTextNode(descDoc, "urn:upnp-org:serviceId:RenderingControl1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "controlURL");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnp/control/RenderingControl1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "eventSubURL");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnp/event/RenderingControl1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "SCPDURL");
  node3 = ixmlDocument_createTextNode(descDoc, "RendererControl.xml");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  ixmlNode_appendChild((IXML_Node*) Ele2, (IXML_Node*) Ele3);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele3 = ixmlDocument_createElement(descDoc, "service");
  Ele4 = ixmlDocument_createElement(descDoc, "serviceType");
  node3 = ixmlDocument_createTextNode(descDoc, "urn:schemas-upnp-org:service:ConnectionManager:1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "serviceId");
  node3 = ixmlDocument_createTextNode(descDoc, "urn:upnp-org:serviceId:ConnectionManager1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "controlURL");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnp/control/ConnectionManager1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "eventSubURL");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnp/event/ConnectionManager1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "SCPDURL");
  node3 = ixmlDocument_createTextNode(descDoc, "ConnectionManager.xml");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  ixmlNode_appendChild((IXML_Node*) Ele2, (IXML_Node*) Ele3);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  Ele3 = ixmlDocument_createElement(descDoc, "service");
  Ele4 = ixmlDocument_createElement(descDoc, "serviceType");
  node3 = ixmlDocument_createTextNode(descDoc, "urn:schemas-upnp-org:service:AVTransport:1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "serviceId");
  node3 = ixmlDocument_createTextNode(descDoc, "urn:upnp-org:serviceId:AVTransport1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "controlURL");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnp/control/AVTransport1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "eventSubURL");
  node3 = ixmlDocument_createTextNode(descDoc, "/upnp/event/AVTransport1");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  Ele4 = ixmlDocument_createElement(descDoc, "SCPDURL");
  node3 = ixmlDocument_createTextNode(descDoc, "AVTransport.xml");
  ixmlNode_appendChild((IXML_Node*) Ele4, node3);
  ixmlNode_appendChild((IXML_Node*) Ele3, (IXML_Node*) Ele4);
  ixmlNode_appendChild((IXML_Node*) Ele2, (IXML_Node*) Ele3);
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

	
  ixmlNode_appendChild((IXML_Node*) Ele, (IXML_Node*) Ele2);

  ixmlNode_appendChild(node, (IXML_Node*) Ele);

  return descDoc;
}

char* cRenderer::get_uuid() 
{
  char* home=getenv("HOME");
  char pfad_to_uuid[100];
  sprintf(pfad_to_uuid,"%s/.renderer/uuid",home);
  // uuid holen, bzw generieren
  char uuid[42];
  std::ifstream uuid_in;
  uuid_in.open(pfad_to_uuid,std::ifstream::in) ;
  if (!uuid_in.is_open()) {
    return generate_uuid();
  }
	
  uuid_in.read(uuid,45);
  if(strlen(uuid)!=41) {
    return generate_uuid();
  }
	
  uuid_in.close();
  cout << "alt: " << uuid << endl;
  return strdup(uuid);
}

char* cRenderer::generate_uuid() 
{
  char* home=getenv("HOME");
  char pfad_to_uuid[100];
  sprintf(pfad_to_uuid,"%s/.renderer/uuid",home);
  //uuid generieren
  char uuid[42];
  uuid_t uu; 
  uuid_generate(uu);
  sprintf(uuid,"uuid:%02x%02x%02x%02x-%02x%02x-%02x%02x-%02x%02x-%02x%02x%02x%02x%02x%02x",
	  uu[0],uu[1],uu[2],uu[3],uu[4],uu[5],uu[6],uu[7],uu[8],uu[9],uu[10],uu[11],uu[12],uu[13],uu[14],uu[15]);
  std::ofstream uuid_out;
  uuid_out.open(pfad_to_uuid,std::ofstream::out);
  if (!uuid_out.is_open())
    cout << "couldn't open uuidfile" << endl;
  uuid_out.write(uuid,strlen(uuid));
  uuid_out.close();
  cout << "neu: " << uuid << endl;
  return strdup(uuid);
}

void cRenderer::trim(string& str)
{
  string::size_type start = str.find_first_not_of(' ');
  string::size_type end = str.find_last_not_of(' ');
  str = str.substr(start == string::npos ? 0 : start, 
		   end == string::npos ? str.length()-1 : end-start+1);
}

void cRenderer::str_replace(string needle, string replacement, string& haystack)
{
  string::size_type start = haystack.find(needle);
  if (start!=string::npos && start>0)
    haystack = haystack.replace( start, needle.length(), replacement);
}

void cRenderer::getVarName(string& varName, string& haystack)
{
  int end = haystack.find_first_of("=");
  if (end != string::npos)
    varName = haystack.substr(0,end);
}

void cRenderer::getValue(string& value, string& haystack)
{
  int begin = haystack.find_first_of("=");
  if (begin != string::npos)
    value = haystack.substr(begin+1);
}

void cRenderer::play()
{
  Simplefile empty;
  audioplayer->set_cur_nr(empty);
  audioplayer->set_streaming(true);

  Simplefile tmp;
  tmp.path = (services[AVTransport])->value[cAVTransport::AV_AVTransportURI];

  audioplayer->addfile(tmp);

  // FIXME: suck metadata to get total time!

  string s = conv::itos(audioplayer->p_total_time());

  (services[AVTransport])->setVar(AVTransport,14,
				  cAVTransport::AV_TransportState,"PLAYING",
				  cAVTransport::AV_CurrentTrack,"1",
				  cAVTransport::AV_NumberOfTracks,"1",
				  cAVTransport::AV_CurrentTrackURI,(services[AVTransport])->value[cAVTransport::AV_AVTransportURI],
				  cAVTransport::AV_CurrentTrackMetaData,"",
				  cAVTransport::AV_AbsoluteTimePosition,(services[AVTransport])->value[cAVTransport::AV_AbsoluteTimePosition],
  				  cAVTransport::AV_CurrentTrackDuration,s.c_str());
}

void cRenderer::stop()
{
  audioplayer->stop();
  audioplayer->set_streaming(false);
}

void cRenderer::pause()
{
  audioplayer->pause();
}

void cRenderer::run()
{
  char cur_time[100];
  while (true) {
    if (audioplayer->is_playing()) {
      string s = conv::itos(audioplayer->p_cur_time());
      seconds2tstamp(cur_time, s.c_str());
    
      (services[AVTransport])->setVar(AVTransport,4,
				      cAVTransport::AV_RelativeTimePosition,cur_time,
				      cAVTransport::AV_CurrentTrackDuration,cur_time);
      sleep(1000); // ms
    } else {
      sleep(100); // ms
    }
  }
}

